
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-07-29T12:07:43-0300",
 "dirty": false,
 "error": null,
 "full-revisionid": "4ba0450a533ee5c9b65ed8ae70df97cfb829ef8d",
 "version": "2.66.2b2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""Add property id to privacy request model

Revision ID: a3c173391603
Revises: 5fe01e730171
Create Date: 2024-06-04 17:40:35.230801

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a3c173391603"
down_revision = "5fe01e730171"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacyrequest", sa.Column("property_id", sa.String(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyrequest", "property_id")
    # ### end Alembic commands ###



from typing import Any

from botocore.exceptions import ClientError
from google.auth.exceptions import GoogleAuthError
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from loguru import logger

from fides.api.models.storage import StorageConfig
from fides.api.schemas.storage.storage import (
    SUPPORTED_STORAGE_SECRETS,
    AWSAuthMethod,
    StorageSecrets,
    StorageType,
)
from fides.api.util.aws_util import get_aws_session


def secrets_are_valid(
    secrets: SUPPORTED_STORAGE_SECRETS,
    storage_config: StorageConfig,
) -> bool:
    """Authenticates upload destination with appropriate upload method"""
    uploader: Any = _get_authenticator_from_config(storage_config.type)  # type: ignore
    return uploader(storage_config, secrets)


def _s3_authenticator(
    config: StorageConfig, secrets: dict[StorageSecrets, Any]
) -> bool:
    """Authenticates secrets for s3, returns true if secrets are valid"""
    try:
        get_aws_session(config.details["auth_method"] or AWSAuthMethod.SECRET_KEYS.value, secrets.model_dump(mode="json"))  # type: ignore
        return True
    except ClientError:
        return False


def _gcs_authenticator(storage_config: StorageConfig, secrets: dict) -> bool:
    """Autenticates secrets for Google Cloud Storage, returns true if secrets are valid"""
    try:
        credentials = service_account.Credentials.from_service_account_info(
            dict(secrets),
            scopes=["https://www.googleapis.com/auth/devstorage.read_only"],
        )
        # To validate the credentials, it is necessary to make a request to Google Cloud API
        credentials.refresh(Request())
        return True

    except GoogleAuthError as auth_error:
        logger.warning(
            "Google authentication error trying to authenticate GCS secrets: {}",
            auth_error,
        )
        return False

    except Exception as e:
        logger.warning("Unexpected error authenticating GCS secrets: {}", e)
        return False


def _get_authenticator_from_config(storage_type: StorageType) -> Any:
    """Determines which uploader method to use based on storage type"""
    return {
        StorageType.s3.value: _s3_authenticator,
        StorageType.gcs.value: _gcs_authenticator,
    }[storage_type.value]



from typing import Any, Callable, List

from pydantic import BaseModel, ConfigDict
from pydantic_core import core_schema


class NoValidationSchema(BaseModel):
    """
    A schema to be used for API documentation only, when validation is
    handled later in the request process, but we still want valid request
    schemas to show up in the docs.
    """

    @classmethod
    def __get_pydantic_core_schema__(  # pylint: disable=arguments-differ
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        """Custom override for Pydantic V2 - allows us to defer validation for
        connection secrets schemas until later."""

        schema = handler(source_type)

        def val(v: Any, _: core_schema.ValidatorFunctionWrapHandler) -> Any:
            return v

        return core_schema.no_info_wrap_validator_function(
            function=val, schema=schema, serialization=schema.get("serialization")
        )


class FidesSchema(BaseModel):
    """A base template for all other Fides Schemas to inherit from."""

    @classmethod
    def get_field_names(cls) -> List[str]:
        """Return a list of all field names specified on this schema."""
        return list(cls.schema().get("properties", {}).keys())

    model_config = ConfigDict(from_attributes=True)



from enum import Enum
from typing import ClassVar, List

from pydantic import Field

from fides.api.custom_types import AnyHttpUrlStringRemovesSlash
from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class PeriodicIntegrationFrequency(Enum):
    """Enum for periodic integration frequency"""

    daily = "daily"
    weekly = "weekly"
    monthly = "monthly"


class DatahubSchema(ConnectionConfigSecretsSchema):
    datahub_server_url: AnyHttpUrlStringRemovesSlash = Field(
        title="DataHub server URL",
        description="The URL of your DataHub server.",
    )
    datahub_token: str = Field(
        title="DataHub token",
        description="The token used to authenticate with your DataHub server.",
        json_schema_extra={"sensitive": True},
    )
    frequency: PeriodicIntegrationFrequency = Field(
        title="Frequency",
        description="The frequency at which the integration should run. Available options are daily, weekly, and monthly.",
    )
    glossary_node: str = Field(
        title="Glossary node",
        description="The glossary node name to use on Datahub for Fides Data Categories. (e.g. FidesDataCategories)",
    )

    _required_components: ClassVar[List[str]] = ["datahub_server_url", "datahub_token"]


class DatahubDocsSchema(DatahubSchema, NoValidationSchema):
    """
    Datahub Schema for API docs.
    """



from __future__ import annotations

import json
from asyncio import sleep
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Set

from httpx import AsyncClient
from loguru import logger
from sqlalchemy import text
from sqlalchemy.sql.elements import TextClause

from fides.api.common_exceptions import PrivacyRequestNotFound
from fides.api.models.privacy_request import (
    EXITED_EXECUTION_LOG_STATUSES,
    PrivacyRequest,
    RequestTask,
)
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.drp_privacy_request import DrpPrivacyRequestCreate
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import (
    PrivacyRequestResponse,
    PrivacyRequestStatus,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.tasks import DSR_QUEUE_NAME, DatabaseTask, celery_app
from fides.api.tasks.scheduled.scheduler import scheduler
from fides.api.util.cache import (
    FidesopsRedis,
    celery_tasks_in_flight,
    get_async_task_tracking_cache_key,
    get_cache,
)
from fides.api.util.lock import redis_lock
from fides.common.api.v1.urn_registry import PRIVACY_REQUESTS, V1_URL_PREFIX
from fides.config import CONFIG

PRIVACY_REQUEST_STATUS_CHANGE_POLL = "privacy_request_status_change_poll"
DSR_DATA_REMOVAL = "dsr_data_removal"
INTERRUPTED_TASK_REQUEUE_POLL = "interrupted_task_requeue_poll"


def build_required_privacy_request_kwargs(
    requested_at: Optional[datetime],
    policy_id: str,
    verification_required: bool,
    authenticated: bool,
) -> Dict[str, Any]:
    """Build kwargs required for creating privacy request

    If identity verification is on, the request will have an initial status of
    "identity_unverified", otherwise, it will have an initial status of "pending".
    """
    status = (
        PrivacyRequestStatus.identity_unverified
        if verification_required and not authenticated
        else PrivacyRequestStatus.pending
    )
    return {
        "requested_at": requested_at,
        "policy_id": policy_id,
        "status": status,
    }


def cache_data(
    privacy_request: PrivacyRequest,
    identity: Identity,
    encryption_key: Optional[str],
    drp_request_body: Optional[DrpPrivacyRequestCreate],
    custom_privacy_request_fields: Optional[Dict[str, Any]] = None,
) -> None:
    """Cache privacy request data"""
    # Store identity and encryption key in the cache
    logger.info("Caching identity for privacy request {}", privacy_request.id)
    privacy_request.cache_identity(identity)
    privacy_request.cache_custom_privacy_request_fields(custom_privacy_request_fields)
    privacy_request.cache_encryption(encryption_key)  # handles None already

    if drp_request_body:
        privacy_request.cache_drp_request_body(drp_request_body)


def get_async_client() -> AsyncClient:
    """Return an async client used to make API requests"""
    return AsyncClient()


async def poll_server_for_completion(
    privacy_request_id: str,
    server_url: str,
    token: str,
    *,
    poll_interval_seconds: int = 30,
    timeout_seconds: int = 1800,  # 30 minutes
    client: AsyncClient | None = None,
) -> PrivacyRequestResponse:
    """Poll a server for privacy request completion.

    Requests will report complete with if they have a status of canceled, complete,
    denied, or error. By default the polling will time out if not completed in 30
    minutes, time can be overridden by setting the timeout_seconds.
    """
    url = (
        f"{server_url}{V1_URL_PREFIX}{PRIVACY_REQUESTS}?request_id={privacy_request_id}"
    )
    start_time = datetime.now()
    elapsed_time = 0.0
    while elapsed_time < timeout_seconds:
        if client:
            response = await client.get(
                url, headers={"Authorization": f"Bearer {token}"}
            )
        else:
            async_client = get_async_client()
            response = await async_client.get(
                url, headers={"Authorization": f"Bearer {token}"}
            )
        response.raise_for_status()

        # Privacy requests are returned paginated. Since this is searching for a specific
        # privacy request there should only be one value present in items.
        items = response.json()["items"]
        if not items:
            raise PrivacyRequestNotFound(
                f"No privacy request found with id '{privacy_request_id}'"
            )
        status = PrivacyRequestResponse(**items[0])
        if status.status and status.status in (
            PrivacyRequestStatus.complete,
            PrivacyRequestStatus.canceled,
            PrivacyRequestStatus.error,
            PrivacyRequestStatus.denied,
        ):
            return status

        await sleep(poll_interval_seconds)
        time_delta = datetime.now() - start_time
        elapsed_time = time_delta.seconds
    raise TimeoutError(
        f"Timeout of {timeout_seconds} seconds has been exceeded while waiting for privacy request {privacy_request_id}"
    )


def initiate_poll_for_exited_privacy_request_tasks() -> None:
    """Initiates scheduler to check if a Privacy Request's status needs to be flipped when all
    Request Tasks have had a chance to run"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add Privacy Request Status Change job."

    logger.info("Initiating scheduler for Privacy Request Status Change")
    scheduler.add_job(
        func=poll_for_exited_privacy_request_tasks,
        trigger="interval",
        kwargs={},
        id=PRIVACY_REQUEST_STATUS_CHANGE_POLL,
        coalesce=True,
        replace_existing=True,
        seconds=CONFIG.execution.state_polling_interval,
    )


@celery_app.task(base=DatabaseTask, bind=True)
def poll_for_exited_privacy_request_tasks(self: DatabaseTask) -> Set[str]:
    """
    Mark a privacy request as errored if all of its Request Tasks have run but some have errored.

    When a Request Task fails, it marks itself and *every Request Task that can be reached by the current
    Request Task* as failed. However, other Request Tasks independent of this path should still have an
    opportunity to run. We wait until everything has run before marking the Privacy Request as errored so it
    can be reprocessed.
    """
    with self.get_new_session() as db:
        logger.debug("Polling for privacy requests awaiting errored status change")

        # Privacy Requests that needed approval are in "Approved" status as they process.
        # Privacy Requests that didn't need approval are "In Processing".
        # Privacy Requests in these states should be examined to see if all of its Request Tasks have had a chance
        # to complete.
        in_progress_privacy_requests = (
            db.query(PrivacyRequest)
            .filter(
                PrivacyRequest.status.in_(
                    [PrivacyRequestStatus.in_processing, PrivacyRequestStatus.approved]
                )
            )
            # Only look at Privacy Requests that haven't been deleted
            .filter(PrivacyRequest.deleted_at.is_(None))
            .order_by(PrivacyRequest.created_at)
        )

        # TODO: With this approach, we're making 3*n queries , where n is the number of in-progress privacy requests.
        # We could optimize this to just get all the RequestTasks for all in-progress privacy requests in one single
        # query and then process them in-memory. This could be more efficient.
        def privacy_request_has_errored_tasks(
            privacy_request: PrivacyRequest, task_type: ActionType
        ) -> bool:
            """Check if a privacy request has exited all its tasks of the given type,
            and at least one of the tasks has errored.
            We specifically only query for the RequestTask.status column to avoid
            querying for the entire RequestTask row.
            """
            tasks_statuses_query = db.query(RequestTask.status).filter(
                RequestTask.privacy_request_id == privacy_request.id,
                RequestTask.action_type == task_type,
            )

            statuses = set(status for status, in tasks_statuses_query.all())
            all_exited = all(
                status in EXITED_EXECUTION_LOG_STATUSES for status in statuses
            )
            if all_exited and ExecutionLogStatus.error in statuses:
                logger.info(
                    f"Marking {task_type.value} step of {privacy_request.id} as error"
                )
                return True

            return False

        marked_as_errored: Set[str] = set()
        for pr in in_progress_privacy_requests.all():
            if pr.consent_tasks.count():
                # Consent propagation tasks - these are not created until access and erasure steps are complete.
                if privacy_request_has_errored_tasks(pr, ActionType.consent):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

            if pr.erasure_tasks.count():
                # Erasure tasks are created at the same time as access tasks but if any are errored, this means
                # we made it to the erasure section
                if privacy_request_has_errored_tasks(pr, ActionType.erasure):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

            if pr.access_tasks.count():
                if privacy_request_has_errored_tasks(pr, ActionType.access):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

        return marked_as_errored


def initiate_scheduled_dsr_data_removal() -> None:
    """Initiates scheduler to cleanup obsolete access and erasure data"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add DSR data removal job."

    logger.info("Initiating scheduler for DSR Data Removal")
    scheduler.add_job(
        func=remove_saved_dsr_data,
        kwargs={},
        id=DSR_DATA_REMOVAL,
        coalesce=False,
        replace_existing=True,
        trigger="cron",
        minute="0",
        hour="2",
        day="*",
        timezone="US/Eastern",
    )


@celery_app.task(base=DatabaseTask, bind=True)
def remove_saved_dsr_data(self: DatabaseTask) -> None:
    """
    Remove saved customer data that is no longer needed to facilitate running the access or erasure request.
    """
    with self.get_new_session() as db:
        logger.info("Running DSR Data Removal Task to cleanup obsolete user data")

        # Remove old request tasks which potentially contain encrypted PII
        remove_dsr_data: TextClause = text(
            """
            DELETE FROM requesttask
            USING privacyrequest
            WHERE requesttask.privacy_request_id = privacyrequest.id
            AND requesttask.created_at < :ttl
            AND privacyrequest.status = 'complete';
            """
        )

        result = db.execute(
            remove_dsr_data,
            {
                "ttl": (
                    datetime.now()
                    - timedelta(seconds=CONFIG.execution.request_task_ttl)
                ),
            },
        )
        affected_rows = result.rowcount
        logger.info(
            f"Deleted {affected_rows} expired request tasks via DSR Data Removal Task."
        )

        # Remove columns from old privacyrequests that potentially contain encrypted PII
        # or URL's that contain encrypted PII.
        remove_data_from_privacy_request: TextClause = text(
            """
            UPDATE privacyrequest
            SET filtered_final_upload = null, access_result_urls = null
            WHERE privacyrequest.updated_at < :ttl
            AND privacyrequest.status = 'complete';
            """
        )

        db.execute(
            remove_data_from_privacy_request,
            {
                "ttl": (  # Using Redis Default TTL Seconds by default
                    datetime.now() - timedelta(seconds=CONFIG.redis.default_ttl_seconds)
                ),
            },
        )

        db.commit()


def initiate_interrupted_task_requeue_poll() -> None:
    """Initiates scheduler to check for and requeue interrupted tasks"""

    if CONFIG.test_mode or not CONFIG.execution.use_dsr_3_0:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add interrupted task requeue job."

    logger.info("Initiating scheduler for interrupted task requeue")
    scheduler.add_job(
        func=requeue_interrupted_tasks,
        trigger="interval",
        kwargs={},
        id=INTERRUPTED_TASK_REQUEUE_POLL,
        coalesce=True,
        replace_existing=True,
        seconds=CONFIG.execution.interrupted_task_requeue_interval,
    )


def get_cached_task_id(entity_id: str) -> Optional[str]:
    """Gets the cached task ID for a privacy request or request task by ID."""
    cache: FidesopsRedis = get_cache()
    task_id = cache.get(get_async_task_tracking_cache_key(entity_id))
    return task_id


REQUEUE_INTERRUPTED_TASKS_LOCK = "requeue_interrupted_tasks_lock"


def _get_task_ids_from_dsr_queue(
    redis_client: FidesopsRedis, chunk_size: int = 100
) -> Set[str]:
    """
    Get all task IDs from a Redis queue in a memory-efficient way.

    Args:
        redis_client: Redis client
        chunk_size: Size of chunks to process

    Returns:
        Set of task IDs
    """
    queued_tasks_ids = set()
    queue_length = redis_client.llen(DSR_QUEUE_NAME)

    # Process the queue in chunks to avoid memory issues with large queues
    for offset in range(0, queue_length, chunk_size):
        # Get a chunk of tasks (at most chunk_size elements)
        tasks_chunk = redis_client.lrange(
            DSR_QUEUE_NAME, offset, offset + chunk_size - 1
        )

        # Extract task IDs from the chunk
        for task in tasks_chunk:
            try:
                queued_tasks_ids.add(json.loads(task)["headers"]["id"])
            except (json.JSONDecodeError, KeyError) as e:
                logger.warning(f"Failed to parse task from queue: {e}")
                continue

    return queued_tasks_ids


# pylint: disable=too-many-branches
@celery_app.task(base=DatabaseTask, bind=True)
def requeue_interrupted_tasks(self: DatabaseTask) -> None:
    """
    Requeue interrupted tasks for privacy requests that are in progress.

    This function checks for privacy requests that are currently pending or in progress and
    verifies if their associated tasks are still in the queue or running. If any
    task is found to be interrupted (i.e., not in the queue and not running), the
    privacy request is requeued to ensure its completion.

    The function performs the following steps:
    1. Retrieves all task IDs currently in the queue.
    2. Fetches all in-progress privacy requests from the database.
    3. Checks each privacy request to determine if its tasks are still active.
    4. Requeues the privacy request if any of its tasks are found to be interrupted.
    """
    redis_conn: FidesopsRedis = get_cache()

    with redis_lock(REQUEUE_INTERRUPTED_TASKS_LOCK, 600) as lock:
        if not lock:
            return
        with self.get_new_session() as db:
            logger.debug("Starting check for interrupted tasks to requeue")

            # Get all in-progress privacy requests that haven't been updated in the last 5 minutes
            in_progress_requests = (
                db.query(PrivacyRequest)
                .filter(
                    PrivacyRequest.status.in_(
                        [
                            PrivacyRequestStatus.in_processing,
                            PrivacyRequestStatus.approved,
                        ]
                    )
                )
                .filter(PrivacyRequest.deleted_at.is_(None))
                .order_by(PrivacyRequest.created_at)
            )

            if not in_progress_requests.count():
                logger.debug("No in-progress privacy requests to check")
                return

            logger.debug(
                f"Found {in_progress_requests.count()} privacy requests to check"
            )

            # Get task IDs from the queue in a memory-efficient way
            queued_tasks_ids = _get_task_ids_from_dsr_queue(redis_conn)

            # Check each privacy request
            for privacy_request in in_progress_requests:
                should_requeue = False
                logger.debug(f"Checking tasks for privacy request {privacy_request.id}")

                task_id = get_cached_task_id(privacy_request.id)

                # If the task ID is not cached, we can't check if it's running
                if not task_id:
                    continue

                # Check if the main privacy request task is active
                if task_id not in queued_tasks_ids and not celery_tasks_in_flight(
                    [task_id]
                ):
                    request_tasks_count = (
                        db.query(RequestTask)
                        .filter(RequestTask.privacy_request_id == privacy_request.id)
                        .count()
                    )
                    if request_tasks_count == 0:
                        logger.warning(
                            f"The task for privacy request {privacy_request.id} was terminated before it could schedule any request tasks, requeueing privacy request"
                        )
                        should_requeue = True

                    request_tasks_in_progress = (
                        db.query(RequestTask.id)
                        .filter(RequestTask.privacy_request_id == privacy_request.id)
                        .filter(
                            RequestTask.status.in_(
                                [
                                    ExecutionLogStatus.in_processing,
                                    ExecutionLogStatus.pending,
                                ]
                            )
                        )
                        .all()
                    )
                    request_task_ids_in_progress = [
                        task[0] for task in request_tasks_in_progress
                    ]

                    # Check each individual request task
                    for request_task_id in request_task_ids_in_progress:
                        subtask_id = get_cached_task_id(request_task_id)

                        # If the task ID is not cached, we can't check if it's running
                        if not subtask_id:
                            continue

                        if (
                            subtask_id not in queued_tasks_ids
                            and not celery_tasks_in_flight([subtask_id])
                        ):
                            logger.warning(
                                f"Request task {request_task_id} is not in the queue or running, requeueing privacy request"
                            )
                            should_requeue = True
                            break

                # Requeue the privacy request if needed
                if should_requeue:
                    from fides.service.privacy_request.privacy_request_service import (  # pylint: disable=cyclic-import
                        PrivacyRequestError,
                        _requeue_privacy_request,
                    )

                    try:
                        _requeue_privacy_request(db, privacy_request)
                    except PrivacyRequestError as exc:
                        logger.error(exc.message)



"""add eu locations migration

Revision ID: d2996381c4dd
Revises: 2be84e68df32
Create Date: 2023-06-20 21:43:45.404454

"""

from typing import Dict, List, Optional

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
from sqlalchemy.engine import Connection, ResultProxy
from sqlalchemy.sql.elements import TextClause

revision = "d2996381c4dd"
down_revision = "2be84e68df32"
branch_labels = None
depends_on = None


region_map: Dict = {
    "eu_be": "be",
    "eu_bg": "bg",
    "eu_cz": "cz",
    "eu_dk": "dk",
    "eu_de": "de",
    "eu_ee": "ee",
    "eu_ie": "ie",
    "eu_el": "gr",
    "eu_es": "es",
    "eu_fr": "fr",
    "eu_hr": "hr",
    "eu_it": "it",
    "eu_cy": "cy",
    "eu_lv": "lv",
    "eu_lt": "lt",
    "eu_lu": "lu",
    "eu_hu": "hu",
    "eu_mt": "mt",
    "eu_nl": "nl",
    "eu_at": "at",
    "eu_pl": "pl",
    "eu_pt": "pt",
    "eu_ro": "ro",
    "eu_si": "si",
    "eu_sk": "sk",
    "eu_fi": "fi",
    "eu_se": "se",
    "isl": "is",
    "nor": "no",
}


def get_replacement_region(
    region: Optional[str], region_mapping: Dict[str, str]
) -> Optional[str]:
    """Get new region from region_mapping if it exists"""
    if not region:
        return region

    return region_mapping.get(region, region)


def upgrade_region_list(
    bind: Connection,
    select_query: TextClause,
    update_query: TextClause,
    region_mapping: Dict[str, str],
) -> None:
    """Upgrade regions for tables that have a "regions" list field"""
    existing_records: ResultProxy = bind.execute(select_query)
    for row in existing_records:
        regions: List[str] = row["regions"] or []
        regions_updated: bool = False
        for i, val in enumerate(regions):
            new_region = get_replacement_region(val, region_mapping)
            if val != new_region:
                regions_updated = True
                regions[i] = new_region

        if regions_updated:
            bind.execute(
                update_query,
                {"id": row["id"], "updated_regions": regions},
            )


def upgrade_region(
    bind: Connection,
    select_query: TextClause,
    update_query: TextClause,
    region_col_label: str,
    region_mapping: Dict[str, str],
):
    """Upgrade regions for tables that have a character varying field that stores a single region"""
    existing_records: ResultProxy = bind.execute(select_query)
    for row in existing_records:
        region: str = row[region_col_label]
        new_region = get_replacement_region(region, region_mapping)

        if region != new_region:
            bind.execute(
                update_query,
                {"id": row["id"], "updated_region": new_region},
            )


def migration_eu_regions(region_mapping: Dict[str, str]) -> None:
    """Generic migration that can upgrade or downgrade regions"""
    bind: Connection = op.get_bind()

    upgrade_region_list(
        bind,
        text("SELECT id, regions FROM privacynotice;"),
        text("UPDATE privacynotice SET regions = :updated_regions WHERE id= :id"),
        region_mapping,
    )

    upgrade_region_list(
        bind,
        text("SELECT id, regions FROM privacynoticetemplate;"),
        text(
            "UPDATE privacynoticetemplate SET regions = :updated_regions WHERE id= :id"
        ),
        region_mapping,
    )

    upgrade_region_list(
        bind,
        text("SELECT id, regions FROM privacynoticehistory;"),
        text(
            "UPDATE privacynoticehistory SET regions = :updated_regions WHERE id= :id"
        ),
        region_mapping,
    )

    upgrade_region(
        bind,
        text("SELECT id, region FROM privacyexperience;"),
        text("UPDATE privacyexperience SET region = :updated_region WHERE id= :id"),
        region_col_label="region",
        region_mapping=region_mapping,
    )

    upgrade_region(
        bind,
        text("SELECT id, user_geography FROM privacypreferencehistory;"),
        text(
            "UPDATE privacypreferencehistory SET user_geography = :updated_region WHERE id= :id"
        ),
        region_col_label="user_geography",
        region_mapping=region_mapping,
    )


def upgrade():
    migration_eu_regions(region_mapping=region_map)


def downgrade():
    migration_eu_regions(
        region_mapping={value: key for key, value in region_map.items()}
    )



"""add is_default column to storage config

Revision ID: 7fd4601ef88b
Revises: d6c6c6555c86
Create Date: 2023-01-26 14:13:22.538719

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.sql.elements import TextClause

# revision identifiers, used by Alembic.
revision = "7fd4601ef88b"
down_revision = "392992c7733a"
branch_labels = None
depends_on = None


def upgrade():
    # schema migration - add a nullable storageconfig.is_default field. We will shortly make it non-nullable
    op.add_column("storageconfig", sa.Column("is_default", sa.Boolean(), nullable=True))
    op.create_index(
        op.f("ix_storageconfig_is_default"),
        "storageconfig",
        ["is_default"],
        unique=False,
    )
    op.create_index(
        "only_one_default_per_type",
        "storageconfig",
        ["type", "is_default"],
        unique=True,
        postgresql_where=sa.text("is_default"),
    )

    # Data migration - automatically populate storageconfig.is_default column
    bind = op.get_bind()
    existing_storageconfigs: TextClause = bind.execute(
        text("SELECT id FROM storageconfig;")
    )
    for row in existing_storageconfigs:
        update_storage_config_query: TextClause = text(
            "UPDATE storageconfig SET is_default = false WHERE id= :storageconfig_id"
        )
        bind.execute(update_storage_config_query, {"storageconfig_id": row["id"]})

    # now follow up to make it not nullable
    op.alter_column(
        "storageconfig", "is_default", existing_type=sa.Boolean(), nullable=False
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "only_one_default_per_type",
        table_name="storageconfig",
        postgresql_where=sa.text("is_default"),
    )
    op.drop_index(op.f("ix_storageconfig_is_default"), table_name="storageconfig")
    op.drop_column("storageconfig", "is_default")
    # ### end Alembic commands ###



from pydantic import Field

from fides.api.schemas.connection_configuration.connection_secrets_base_aws import (
    BaseAWSSchema,
)


class BaseRDSSchema(BaseAWSSchema):
    """
    Schema to validate the secrets needed to connect to a RDS Database
    """

    db_username: str = Field(
        default="fides_service_user",
        title="DB Username",
        description="The user account used to authenticate and access the databases.",
    )
    region: str = Field(
        title="Region",
        description="The AWS region where the RDS instances are located.",
    )



from __future__ import annotations

from uuid import uuid4

from sqlalchemy import Column, ForeignKey, String

from fides.api.db.base_class import Base


class ExperienceNotices(Base):
    """
    A many-to-many table that links Privacy Notices to shared Privacy Experience Configs.
    """

    def generate_uuid(self) -> str:
        """
        Generates a uuid with a prefix based on the tablename to be used as the
        record's ID value
        """
        try:
            prefix = f"{self.current_column.table.name[:3]}_"  # type: ignore
        except AttributeError:
            prefix = ""
        uuid = str(uuid4())
        return f"{prefix}{uuid}"

    # Overrides Base.id so this is not a primary key.
    # Instead, we have a composite PK of notice_id and experience_config_id
    id = Column(String(255), default=generate_uuid)

    notice_id = Column(
        String,
        ForeignKey("privacynotice.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
        primary_key=True,
    )
    experience_config_id = Column(
        String,
        ForeignKey("privacyexperienceconfig.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
        primary_key=True,
    )



import hashlib
import secrets
from base64 import b64decode, b64encode
from binascii import Error

import bcrypt


def decode_password(password: str) -> str:
    """Tries to decode the string as base64 encoded.

    If not successful it is assumed to be unencoded and returned as it was sent.
    """
    try:
        return b64_str_to_str(password)
    except (Error, UnicodeDecodeError):
        return password


def hash_credential_with_salt(text: bytes, salt: bytes) -> str:
    """
    Hashes the text using bcrypt with the provided salt and returns the hex string representation.

    **This is a computationally expensive hash that should only be used to hash passwords or other
    credential information. For general data hashing use `hash_value_with_salt`.**
    """
    return bcrypt.hashpw(text, salt).hex()


def hash_value_with_salt(text: bytes, salt: bytes) -> str:
    """
    Hashes the text using SHA-256 with the provided salt and returns the hex string representation.

    **This is a quick hashing function only suitable for general data, use `hash_credential_with_salt` to hash
    passwords or other credentials.**
    """
    return hashlib.sha256(text + salt).hexdigest()


def generate_secure_random_string(length: int) -> str:
    """Generates a securely random string using Python secrets library
    that is twice the length of the specified input"""
    return secrets.token_hex(length)


def generate_salt(encoding: str = "UTF-8") -> str:
    """Generates a salt using bcrypt and returns a string using the configured
    default encoding
    """
    return bcrypt.gensalt().decode(encoding)


def bytes_to_b64_str(bytestring: bytes, encoding: str = "UTF-8") -> str:
    """Converts random bytes into a utf-8 encoded string"""
    return b64encode(bytestring).decode(encoding)


def b64_str_to_bytes(encoded_str: str, encoding: str = "UTF-8") -> bytes:
    """Converts encoded string into bytes"""
    return b64decode(encoded_str.encode(encoding))


def b64_str_to_str(encoded_str: str, encoding: str = "UTF-8") -> str:
    """Converts encoded string into str"""
    return b64decode(encoded_str).decode(encoding)


def str_to_b64_str(string: str, encoding: str = "UTF-8") -> str:
    """Converts str into a utf-8 encoded string"""
    return b64encode(string.encode(encoding)).decode(encoding)



from typing import List, Optional, Type

from fides.api.schemas.masking.masking_configuration import PreserveMaskingConfiguration
from fides.api.schemas.masking.masking_strategy_description import (
    MaskingStrategyDescription,
)
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy


class PreserveMaskingStrategy(MaskingStrategy):
    """Preserves provided values as-is without any modifications."""

    name = "preserve"
    configuration_model = PreserveMaskingConfiguration

    def __init__(
        self,
        configuration: PreserveMaskingConfiguration,
    ):
        """For parity with other MaskingStrategies, but for PreserveMaskingStrategy, nothing is pulled from the config"""

    def mask(
        self, values: Optional[List[str]], request_id: Optional[str]
    ) -> Optional[List[str]]:
        """Returns the original values without any modifications"""
        return values

    def secrets_required(self) -> bool:
        return False

    @classmethod
    def get_description(cls: Type[MaskingStrategy]) -> MaskingStrategyDescription:
        return MaskingStrategyDescription(
            name=cls.name,
            description="Preserves the input values as-is without any modifications",
            configurations=[],
        )

    @staticmethod
    def data_type_supported(data_type: Optional[str]) -> bool:
        """Determines whether or not the given data type is supported by this masking strategy"""
        return True



from typing import Optional
from zipfile import ZipFile

from defusedxml.ElementTree import fromstring
from tinycss2 import parse_stylesheet  # type: ignore
from tinycss2.ast import FunctionBlock  # type: ignore

from fides.api.common_exceptions import ValidationError

MAX_FILE_SIZE = 16 * 1024 * 1024  # 16 MB
CHUNK_SIZE = 1024


def verify_svg(contents: str) -> None:
    """
    Verifies the provided SVG content.

    This function checks the given SVG content string for potential issues and throws an exception if any are found.
    It first attempts to parse the SVG content using 'defusedxml.fromstring'. If the parsing is unsuccessful, this
    will raise an exception, indicating that the SVG content may contain unsafe XML.

    :param contents: The SVG content as a string.
    :raises ValidationError: If the SVG content contains unsafe XML or 'use xlink'
    """
    try:
        fromstring(contents)
    except Exception:
        raise ValidationError("SVG file contains unsafe XML.")

    if "use xlink" in contents:
        raise ValidationError("SVG files with xlink references are not allowed.")


def verify_zip(zip_file: ZipFile, max_file_size: Optional[int] = None) -> None:
    """
    Function to safely verify the contents of zipped files. It prevents potential
    'zip bomb' attacks by checking the file size of the files in the zip without fully
    extracting them. If the size of any file in the zip exceeds the specified
    max_file_size, it raises a ValueError. If the max_file_size is not provided,
    it uses a default value of 16 MB.

    :param zip_file: A ZipFile object to be verified.
    :param max_file_size: An optional integer specifying the maximum bytes allowed per file. If not provided, a default value is used.
    :raises ValueError: If a file in the zip file exceeds the maximum allowed size
    """

    if max_file_size is None:
        max_file_size = MAX_FILE_SIZE

    for file_info in zip_file.infolist():
        file_size = 0

        with zip_file.open(file_info) as file:
            # wraps the file read in an iterator that stops once no bytes
            # are returned or the max file size is reached
            for chunk in iter(lambda f=file: f.read(CHUNK_SIZE), b""):  # type: ignore
                file_size += len(chunk)

                if file_size > max_file_size:
                    raise ValueError("File size exceeds maximum allowed size")


def verify_css(contents: str) -> None:
    """
    Ensures that the contents are a valid CSS stylesheet. The validation checks for parsing errors,
    import statements, and calls to the url() function.
    """

    nodes = parse_stylesheet(contents, skip_comments=True, skip_whitespace=True)
    for node in nodes:
        # check for parsing errors
        if node.type == "error":
            raise ValidationError("Unable to parse CSS")
        # Check for @import rules
        if node.type == "at-rule" and node.lower_at_keyword == "import":
            raise ValidationError("@import statements are not allowed in the CSS")

        # Check for url() functions
        if node.type == "qualified-rule":
            if any(
                isinstance(token, FunctionBlock) and token.name == "url"
                for token in node.content
            ):
                raise ValidationError("Calls to url() are not allowed in the CSS")



from typing import ClassVar, List, Optional

from pydantic import Field

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class MariaDBSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to a MariaDB Database"""

    host: str = Field(
        title="Host",
        description="The hostname or IP address of the server where the database is running.",
    )
    port: int = Field(
        3306,
        title="Port",
        description="The network port number on which the server is listening for incoming connections (default: 3306).",
    )
    username: Optional[str] = Field(
        None,
        title="Username",
        description="The user account used to authenticate and access the database.",
    )
    password: Optional[str] = Field(
        None,
        title="Password",
        description="The password used to authenticate and access the database.",
        json_schema_extra={"sensitive": True},
    )
    dbname: str = Field(
        description="The name of the specific database within the database server that you want to connect to.",
        title="Database",
    )

    _required_components: ClassVar[List[str]] = ["host", "dbname"]


class MariaDBDocsSchema(MariaDBSchema, NoValidationSchema):
    """MariaDB Secrets Schema for API Docs"""



from functools import partial
from typing import Dict, List, Optional, Type, Union

from fastapi import Depends, HTTPException, Query, Response, Security
from fastapi.concurrency import run_in_threadpool
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi_pagination import Page, Params
from fastapi_pagination.ext.async_sqlalchemy import paginate as async_paginate
from fideslang.models import Dataset as FideslangDataset
from pydantic import ValidationError as PydanticValidationError
from sqlalchemy import not_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session, load_only
from starlette import status
from starlette.status import (
    HTTP_200_OK,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api.deps import get_db
from fides.api.common_exceptions import KeyOrNameAlreadyExists, ValidationError
from fides.api.db.base_class import get_key_from_data
from fides.api.db.crud import list_resource_query
from fides.api.db.ctl_session import get_async_db
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.dataset import DatasetResponse
from fides.api.schemas.filter_params import FilterParams
from fides.api.schemas.taxonomy_extensions import (
    DataCategory,
    DataCategoryCreateOrUpdate,
    DataSubject,
    DataSubjectCreateOrUpdate,
    DataUse,
    DataUseCreateOrUpdate,
)
from fides.api.service.deps import get_dataset_service
from fides.api.util.api_router import APIRouter
from fides.api.util.errors import FidesError, ForbiddenIsDefaultTaxonomyError
from fides.api.util.filter_utils import apply_filters_to_query
from fides.common.api.scope_registry import (
    CTL_DATASET_CREATE,
    CTL_DATASET_DELETE,
    CTL_DATASET_READ,
    CTL_DATASET_UPDATE,
    DATA_CATEGORY_CREATE,
    DATA_CATEGORY_UPDATE,
    DATA_SUBJECT_CREATE,
    DATA_USE_CREATE,
    DATA_USE_UPDATE,
)
from fides.common.api.v1.urn_registry import DATASETS_CLEAN, V1_URL_PREFIX
from fides.service.dataset.dataset_service import (
    DatasetNotFoundException,
    DatasetService,
)

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    Dataset as CtlDataset,
)

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    DataCategory as DataCategoryDbModel,
    DataSubject as DataSubjectDbModel,
    DataUse as DataUseDbModel,
    ModelWithDefaultField,
)

# We create routers to override specific methods in those defined in generic.py
# when we need more custom implementations for only some of the methods in a router.

dataset_router = APIRouter(tags=["Dataset"], prefix=V1_URL_PREFIX)
data_use_router = APIRouter(tags=["DataUse"], prefix=V1_URL_PREFIX)
data_category_router = APIRouter(tags=["DataCategory"], prefix=V1_URL_PREFIX)
data_subject_router = APIRouter(tags=["DataSubject"], prefix=V1_URL_PREFIX)


@dataset_router.post(
    "/dataset",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_CREATE])],
    response_model=FideslangDataset,
    status_code=status.HTTP_201_CREATED,
    name="Create dataset",
)
async def create_dataset(
    dataset: FideslangDataset,
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> Dict:
    """Create a new dataset"""
    try:
        return dataset_service.create_dataset(dataset)
    except (ValidationError, PydanticValidationError) as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(
                exc.errors(include_url=False, include_input=False)
                if isinstance(exc, PydanticValidationError)
                else exc.message
            ),
        )
    except KeyOrNameAlreadyExists as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=str(exc),
        )


@dataset_router.put(
    "/dataset",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_UPDATE])],
    response_model=FideslangDataset,
    status_code=status.HTTP_200_OK,
    name="Update dataset",
)
async def update_dataset(
    dataset: FideslangDataset,
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> Dict:
    """Update an existing dataset"""
    try:
        return dataset_service.update_dataset(dataset)
    except (ValidationError, PydanticValidationError) as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(
                exc.errors(include_url=False, include_input=False)
                if isinstance(exc, PydanticValidationError)
                else exc.message
            ),
        )
    except DatasetNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail={"message": str(e)},
        )


@dataset_router.get(
    "/dataset",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_READ])],
    response_model=Union[Page[DatasetResponse], List[DatasetResponse]],
    name="List datasets (optionally paginated)",
)
async def list_dataset_paginated(
    db: AsyncSession = Depends(get_async_db),
    size: Optional[int] = Query(None, ge=1, le=100),
    page: Optional[int] = Query(None, ge=1),
    search: Optional[str] = Query(None),
    data_categories: Optional[List[str]] = Query(None),
    exclude_saas_datasets: Optional[bool] = Query(False),
    only_unlinked_datasets: Optional[bool] = Query(False),
    connection_type: Optional[ConnectionType] = Query(None),
    minimal: Optional[bool] = Query(False),
) -> Union[Page[DatasetResponse], List[DatasetResponse]]:
    """
    Get a list of all of the Datasets.
    If any pagination parameters (size or page) are provided, then the response will be paginated.
    Otherwise all Datasets will be returned (this may be a slow operation if there are many datasets,
    so using the pagination parameters is recommended).
    Provided filters (search, data_categories, exclude_saas_datasets, only_unlinked_datasets) will be applied,
    returning only the datasets that match ALL of the filters.
    """

    query = select(CtlDataset)

    if minimal:
        # .options() allows us to modify how the query loads data by configuring the query's loading behavior
        # load_only() optimizes the query by only loading the specified columns from the database
        # This reduces memory usage and query time by not loading unnecessary columns
        # The columns specified below are the minimal set needed for the DatasetResponse model
        query = query.options(  # type: ignore[attr-defined]
            load_only(
                CtlDataset.id,
                CtlDataset.fides_key,
                CtlDataset.organization_fides_key,
                CtlDataset.name,
                CtlDataset.created_at,
                CtlDataset.updated_at,
                CtlDataset.meta,
                CtlDataset.fides_meta,
                CtlDataset.description,
            )
        )

    # Add filters for search and data categories
    filter_params = FilterParams(search=search, data_categories=data_categories)
    filtered_query = apply_filters_to_query(
        query=query,
        search_model=CtlDataset,
        taxonomy_model=CtlDataset,
        filter_params=filter_params,
    )

    # If applicable, filter by connection type
    if connection_type:
        filtered_query = filtered_query.where(
            CtlDataset.fides_meta["namespace"]["connection_type"].as_string()
            == connection_type.value
        )

    # If applicable, keep only unlinked datasets
    if only_unlinked_datasets:
        linked_datasets = select([DatasetConfig.ctl_dataset_id])
        filtered_query = filtered_query.where(not_(CtlDataset.id.in_(linked_datasets)))

    # If applicable, remove saas config datasets
    if exclude_saas_datasets:
        saas_subquery = (
            select([ConnectionConfig.saas_config["fides_key"].astext])
            .select_from(ConnectionConfig)  # type: ignore[arg-type]
            .where(ConnectionConfig.saas_config.is_not(None))  # type: ignore[attr-defined]
        )
        filtered_query = filtered_query.where(
            not_(CtlDataset.fides_key.in_(saas_subquery))
        )

    if not page and not size:
        results = await list_resource_query(db, filtered_query, CtlDataset)
        response = [
            DatasetResponse.model_validate(result.__dict__) for result in results
        ]
        return response

    pagination_params = Params(page=page or 1, size=size or 50)
    results = await async_paginate(db, filtered_query, pagination_params)

    validated_items = []
    for result in results.items:  # type: ignore[attr-defined]
        # run pydantic validation in a threadpool to avoid blocking the main thread
        validated_item = await run_in_threadpool(
            partial(DatasetResponse.model_validate, result.__dict__)
        )
        validated_items.append(validated_item)

    results.items = validated_items  # type: ignore[attr-defined]

    return results


@dataset_router.get(
    "/dataset/{fides_key}",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_READ])],
    response_model=FideslangDataset,
    name="Get dataset",
)
async def get_dataset(
    fides_key: str,
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> Dict:
    """Get a single dataset by fides key"""
    try:
        return dataset_service.get_dataset(fides_key)
    except DatasetNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail={"message": str(e)},
        )


@dataset_router.delete(
    "/dataset/{fides_key}",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_DELETE])],
    name="Delete dataset",
)
async def delete_dataset(
    fides_key: str,
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> JSONResponse:
    """Delete a dataset by fides key"""
    try:
        dataset = dataset_service.delete_dataset(fides_key)
        return JSONResponse(
            content={
                "message": "resource deleted",
                "resource": FideslangDataset.model_validate(dataset).model_dump(
                    mode="json"
                ),
            },
        )
    except DatasetNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail={"message": str(e)},
        )


@dataset_router.post(
    "/dataset/upsert",
    dependencies=[
        Security(
            verify_oauth_client,
            scopes=[CTL_DATASET_CREATE, CTL_DATASET_UPDATE],
        )
    ],
)
async def upsert_datasets(
    datasets: List[FideslangDataset],
    response: Response,
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> Dict:
    """
    For any dataset in `datasets` that already exists in the database,
    update the dataset by its `fides_key`. Otherwise, create a new dataset.

    Responds with a `201 Created` if even a single dataset in `datasets`
    did not previously exist. Otherwise, responds with a `200 OK`.
    """
    try:
        inserted, updated = dataset_service.upsert_datasets(datasets)

        response.status_code = (
            status.HTTP_201_CREATED if inserted > 0 else response.status_code
        )

        return {
            "message": f"Upserted {len(datasets)} Dataset(s)",
            "inserted": inserted,
            "updated": updated,
        }
    except (ValidationError, PydanticValidationError) as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(
                exc.errors(include_url=False, include_input=False)
                if isinstance(exc, PydanticValidationError)
                else exc.message
            ),
        )


@dataset_router.get(
    DATASETS_CLEAN,
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_READ])],
    response_model=List[FideslangDataset],
    deprecated=True,
)
def clean_datasets(
    dataset_service: DatasetService = Depends(get_dataset_service),
) -> JSONResponse:
    """
    Clean up names of datasets and upsert them.
    """

    succeeded, failed = dataset_service.clean_datasets()
    return JSONResponse(
        status_code=HTTP_200_OK,
        content={
            "succeeded": succeeded,
            "failed": failed,
        },
    )


def activate_taxonomy_parents(
    resource: Union[DataCategoryDbModel, DataUseDbModel, DataSubjectDbModel],
    db: Session,
) -> None:
    """
    Activates parents to match newly-active taxonomy node.
    """
    parent = resource.parent
    if parent:
        parent.active = True
        db.commit()

        activate_taxonomy_parents(parent, db)


def deactivate_taxonomy_node_and_descendants(
    resource: Union[DataCategoryDbModel, DataUseDbModel, DataSubjectDbModel],
    db: Session,
) -> None:
    """
    Recursively de-activates all descendants of a given taxonomy node.
    """
    resource.active = False
    db.commit()
    children = resource.children

    for child in children:
        # Deactivate current child
        child.active = False
        db.commit()

        # Recursively deactivate all descendants of this child
        deactivate_taxonomy_node_and_descendants(child, db)


def validate_and_create_taxonomy(
    db: Session,
    model: Union[
        Type[DataCategoryDbModel], Type[DataUseDbModel], Type[DataSubjectDbModel]
    ],
    validation_schema: type,
    data: Union[
        DataCategoryCreateOrUpdate, DataUseCreateOrUpdate, DataSubjectCreateOrUpdate
    ],
) -> Dict:
    """
    Validate and create a taxonomy element.
    """
    if not data.fides_key:
        raise FidesError(f"Fides key is required to create a {model.__name__} resource")
    if isinstance(data, ModelWithDefaultField) and data.is_default:
        raise ForbiddenIsDefaultTaxonomyError(
            model.__name__, data.fides_key, action="create"
        )
    validated_taxonomy = validation_schema(**data.model_dump(mode="json"))
    return model.create(db=db, data=validated_taxonomy.model_dump(mode="json"))


def validate_and_update_taxonomy(
    db: Session,
    resource: Union[DataCategoryDbModel, DataUseDbModel, DataSubjectDbModel],
    validation_schema: type,
    data: Union[
        DataCategoryCreateOrUpdate, DataUseCreateOrUpdate, DataSubjectCreateOrUpdate
    ],
) -> Dict:
    """
    Validate and update a taxonomy element.
    """
    if (
        isinstance(data, ModelWithDefaultField)
        and data.is_default != resource.is_default
    ):
        raise ForbiddenIsDefaultTaxonomyError(
            "resource", data.fides_key, action="modify"
        )

    # If active field is being updated, cascade change either up or down
    if hasattr(data, "active"):
        if data.active:
            activate_taxonomy_parents(resource, db)
        else:
            # Cascade down - deactivate current node and children to match newly-deactivated taxonomy item
            deactivate_taxonomy_node_and_descendants(resource, db)

    validated_taxonomy = validation_schema(**data.model_dump(mode="json"))
    return resource.update(db=db, data=validated_taxonomy.model_dump(mode="json"))


def create_or_update_taxonomy(
    db: Session,
    data: Union[
        DataCategoryCreateOrUpdate, DataUseCreateOrUpdate, DataSubjectCreateOrUpdate
    ],
    model: Union[
        Type[DataCategoryDbModel], Type[DataUseDbModel], Type[DataSubjectDbModel]
    ],
    validation_schema: type,
) -> Dict:
    """
    Create or update a taxonomy element.
    If the element is deactivated, it will be updated and re-activated, along with its parents.
    """
    if data.fides_key is None:
        disabled_resource_with_name = (
            db.query(model)
            .filter(
                model.active.is_(False),
                model.name == data.name,
            )
            .first()
        )
        data.fides_key = get_key_from_data(
            {"key": data.fides_key, "name": data.name}, validation_schema.__name__
        )
        if data.parent_key if hasattr(data, "parent_key") else None:
            # Updates fides_key if it is not the root level taxonomy node
            data.fides_key = f"{data.parent_key}.{data.fides_key}"  # type: ignore[union-attr]
        if disabled_resource_with_name:
            data.active = True
            activate_taxonomy_parents(disabled_resource_with_name, db)
            return validate_and_update_taxonomy(
                db, disabled_resource_with_name, validation_schema, data
            )
        return validate_and_create_taxonomy(db, model, validation_schema, data)

    return validate_and_create_taxonomy(db, model, validation_schema, data)


@data_use_router.post(
    "/data_use",
    dependencies=[Security(verify_oauth_client, scopes=[DATA_USE_CREATE])],
    response_model=DataUse,
    status_code=status.HTTP_201_CREATED,
    name="Create",
)
async def create_data_use(
    data_use: DataUseCreateOrUpdate,
    db: Session = Depends(get_db),
) -> Dict:
    """
    Create a data use. Updates existing data use if data use with name already exists and is disabled.
    """
    try:
        return create_or_update_taxonomy(db, data_use, DataUseDbModel, DataUse)
    except KeyOrNameAlreadyExists:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Data use with key {data_use.fides_key} or name {data_use.name} already exists.",
        )
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=str(e),
        )


@data_category_router.post(
    "/data_category",
    dependencies=[Security(verify_oauth_client, scopes=[DATA_CATEGORY_CREATE])],
    response_model=DataCategory,
    status_code=status.HTTP_201_CREATED,
    name="Create",
)
async def create_data_category(
    data_category: DataCategoryCreateOrUpdate,
    db: Session = Depends(get_db),
) -> Dict:
    """
    Create a data category
    """

    try:
        return create_or_update_taxonomy(
            db, data_category, DataCategoryDbModel, DataCategory
        )
    except KeyOrNameAlreadyExists:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Data category with key {data_category.fides_key} or name {data_category.name} already exists.",
        )


@data_subject_router.post(
    "/data_subject",
    dependencies=[Security(verify_oauth_client, scopes=[DATA_SUBJECT_CREATE])],
    response_model=DataSubject,
    status_code=status.HTTP_201_CREATED,
    name="Create",
)
async def create_data_subject(
    data_subject: DataSubjectCreateOrUpdate,
    db: Session = Depends(get_db),
) -> Dict:
    """
    Create a data subject
    """

    try:
        return create_or_update_taxonomy(
            db, data_subject, DataSubjectDbModel, DataSubject
        )
    except KeyOrNameAlreadyExists:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Data subject with key {data_subject.fides_key} or name {data_subject.name} already exists.",
        )


@data_use_router.put(
    "/data_use",
    dependencies=[Security(verify_oauth_client, scopes=[DATA_USE_UPDATE])],
    response_model=DataUse,
    status_code=status.HTTP_200_OK,
    name="Update",
)
async def update_data_use(
    data_use: DataUseCreateOrUpdate,
    db: Session = Depends(get_db),
) -> Dict:
    """
    Update a data use. Ensures updates to "active" are appropriately cascaded.
    """

    resource = DataUseDbModel.get_by(db, field="fides_key", value=data_use.fides_key)
    if not resource:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Data use not found with key: {data_use.fides_key}",
        )
    return validate_and_update_taxonomy(db, resource, DataUse, data_use)


@data_category_router.put(
    "/data_category",
    dependencies=[Security(verify_oauth_client, scopes=[DATA_CATEGORY_UPDATE])],
    response_model=DataCategory,
    status_code=status.HTTP_200_OK,
    name="Update",
)
async def update_data_category(
    data_category: DataCategoryCreateOrUpdate,
    db: Session = Depends(get_db),
) -> Dict:
    """
    Update a data category. Ensures updates to "active" are appropriately cascaded.
    """

    resource = DataCategoryDbModel.get_by(
        db, field="fides_key", value=data_category.fides_key
    )
    if not resource:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Data category not found with key: {data_category.fides_key}",
        )
    return validate_and_update_taxonomy(db, resource, DataCategory, data_category)


GENERIC_OVERRIDES_ROUTER = APIRouter()
GENERIC_OVERRIDES_ROUTER.include_router(dataset_router)
GENERIC_OVERRIDES_ROUTER.include_router(data_use_router)
GENERIC_OVERRIDES_ROUTER.include_router(data_category_router)
GENERIC_OVERRIDES_ROUTER.include_router(data_subject_router)



"""Data use unique together

Revision ID: c4df5d585029
Revises: cf88efa1ad89
Create Date: 2022-09-26 23:12:00.816657

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c4df5d585029"
down_revision = "cf88efa1ad89"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("consent_data_use_key", "consent", type_="unique")
    op.create_unique_constraint(
        "uix_identity_data_use", "consent", ["provided_identity_id", "data_use"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uix_identity_data_use", "consent", type_="unique")
    op.create_unique_constraint("consent_data_use_key", "consent", ["data_use"])
    # ### end Alembic commands ###



import re
from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import ConfigDict, EmailStr, field_validator

from fides.api.cryptography.cryptographic_util import decode_password
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.oauth import AccessToken


class PrivacyRequestReviewer(FidesSchema):
    """Data we can expose via the PrivacyRequest.reviewer relation"""

    id: str
    username: str


class UserCreate(FidesSchema):
    """Data required to create a FidesUser."""

    username: str
    password: Optional[str] = None
    email_address: EmailStr
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    disabled: bool = False

    model_config = ConfigDict(extra="ignore")

    @field_validator("username")
    @classmethod
    def validate_username(cls, username: str) -> str:
        """Ensure username does not have spaces."""
        if " " in username:
            raise ValueError("Usernames cannot have spaces.")
        return username

    @field_validator("password")
    @classmethod
    def validate_password_field(cls, password: str) -> str:
        """Add some password requirements"""
        decoded_password = decode_password(password)
        return UserCreate.validate_password(decoded_password)

    @staticmethod
    def validate_password(password: str) -> str:
        """
        Validate password requirements.
            Raises:
                ValueError: If password does not meet requirements
            Returns:
                str: password
        """
        if len(password) < 8:
            raise ValueError("Password must have at least eight characters.")
        if re.search(r"[\d]", password) is None:
            raise ValueError("Password must have at least one number.")
        if re.search("[A-Z]", password) is None:
            raise ValueError("Password must have at least one capital letter.")
        if re.search("[a-z]", password) is None:
            raise ValueError("Password must have at least one lowercase letter.")
        if re.search(r"[\W_]", password) is None:
            raise ValueError("Password must have at least one symbol.")

        return password


class UserCreateResponse(FidesSchema):
    """Response after creating a FidesUser"""

    id: str


class UserLogin(FidesSchema):
    """Similar to UserCreate except we do not need the extra validation on
    username and password.
    """

    username: str
    password: str

    @field_validator("password")
    @classmethod
    def validate_password(cls, password: str) -> str:
        """Convert b64 encoded password to normal string"""
        return decode_password(password)


class UserResponse(FidesSchema):
    """Response after requesting a User"""

    id: str
    username: str
    created_at: datetime
    email_address: Optional[EmailStr]
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    disabled: Optional[bool] = False
    disabled_reason: Optional[str] = None


class UserLoginResponse(FidesSchema):
    """Similar to UserResponse except with an access token"""

    user_data: UserResponse
    token_data: AccessToken


class UserPasswordReset(FidesSchema):
    """Contains both old and new passwords when resetting a password"""

    old_password: str
    new_password: str

    @field_validator("new_password")
    @classmethod
    def validate_new_password(cls, password: str) -> str:
        """Add some password requirements"""
        decoded_password = decode_password(password)
        return UserCreate.validate_password(decoded_password)


class UserForcePasswordReset(FidesSchema):
    """Only a new password, for the case where the user does not remember their password"""

    new_password: str

    @field_validator("new_password")
    @classmethod
    def validate_new_password(cls, password: str) -> str:
        """Add some password requirements"""
        decoded_password = decode_password(password)
        return UserCreate.validate_password(decoded_password)


class UserUpdate(FidesSchema):
    """Data required to update a FidesUser"""

    email_address: Optional[EmailStr] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None

    model_config = ConfigDict(extra="ignore")


class DisabledReason(Enum):
    """Reasons for why a user is disabled"""

    pending_invite = "pending_invite"



from enum import Enum


class Event(str, Enum):
    """Enum to hold analytics event names"""

    server_start = "server_start"
    endpoint_call = "endpoint_call"


class ExtraData(str, Enum):
    """Enum to hold keys for extra data"""

    fides_source = "fides_source"



"""privacy experiences

Revision ID: e92da354691e
Revises: 5b03859e51b5
Create Date: 2023-04-24 14:49:37.588144

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e92da354691e"
down_revision = "5b03859e51b5"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "privacyexperiencetemplate",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("delivery_mechanism", sa.String(), nullable=False),
        sa.Column("regions", sa.ARRAY(sa.String()), nullable=False),
        sa.Column("component_title", sa.String(), nullable=True),
        sa.Column("component_description", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("link_label", sa.String(), nullable=True),
        sa.Column("confirmation_button_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("acknowledgement_button_label", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyexperiencetemplate_id"),
        "privacyexperiencetemplate",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperiencetemplate_regions"),
        "privacyexperiencetemplate",
        ["regions"],
        unique=False,
    )
    op.create_table(
        "privacyexperience",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("delivery_mechanism", sa.String(), nullable=False),
        sa.Column("regions", sa.ARRAY(sa.String()), nullable=False),
        sa.Column("component_title", sa.String(), nullable=True),
        sa.Column("component_description", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("link_label", sa.String(), nullable=True),
        sa.Column("confirmation_button_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("acknowledgement_button_label", sa.String(), nullable=True),
        sa.Column("version", sa.Float(), nullable=False),
        sa.Column("privacy_experience_template_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["privacy_experience_template_id"],
            ["privacyexperiencetemplate.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyexperience_id"), "privacyexperience", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_privacyexperience_regions"),
        "privacyexperience",
        ["regions"],
        unique=False,
    )
    op.create_table(
        "privacyexperiencehistory",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("delivery_mechanism", sa.String(), nullable=False),
        sa.Column("regions", sa.ARRAY(sa.String()), nullable=False),
        sa.Column("component_title", sa.String(), nullable=True),
        sa.Column("component_description", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("link_label", sa.String(), nullable=True),
        sa.Column("confirmation_button_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("acknowledgement_button_label", sa.String(), nullable=True),
        sa.Column("version", sa.Float(), nullable=False),
        sa.Column("privacy_experience_template_id", sa.String(), nullable=True),
        sa.Column("privacy_experience_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_experience_id"],
            ["privacyexperience.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_experience_template_id"],
            ["privacyexperiencetemplate.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_id"),
        "privacyexperiencehistory",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_regions"),
        "privacyexperiencehistory",
        ["regions"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_privacyexperiencehistory_regions"),
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        op.f("ix_privacyexperiencehistory_id"), table_name="privacyexperiencehistory"
    )
    op.drop_table("privacyexperiencehistory")
    op.drop_index(op.f("ix_privacyexperience_regions"), table_name="privacyexperience")
    op.drop_index(op.f("ix_privacyexperience_id"), table_name="privacyexperience")
    op.drop_table("privacyexperience")
    op.drop_index(
        op.f("ix_privacyexperiencetemplate_regions"),
        table_name="privacyexperiencetemplate",
    )
    op.drop_index(
        op.f("ix_privacyexperiencetemplate_id"), table_name="privacyexperiencetemplate"
    )
    op.drop_table("privacyexperiencetemplate")



"""add manual connector type, and add "paused" status to execution logs

Revision ID: 3a7c5fb119c9
Revises: 5078badb90b9
Create Date: 2022-05-13 19:18:45.400669

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3a7c5fb119c9"
down_revision = "5078badb90b9"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")
    op.execute("alter type executionlogstatus add value 'paused'")


def downgrade():
    # Downgrade connection_type
    op.execute("delete from connectionconfig where connection_type in ('manual')")
    op.execute("delete from executionlog where status = 'paused'")

    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Downgrade executionlogstatus
    op.execute("alter type executionlogstatus rename to executionlogstatus_old")
    op.execute(
        "create type executionlogstatus as enum('in_processing', 'pending', 'complete', 'error', 'retrying')"
    )
    op.execute(
        (
            "alter table executionlog alter column status type executionlogstatus using "
            "status::text::executionlogstatus"
        )
    )
    op.execute("drop type executionlogstatus_old")



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Optional, Set, Tuple

from fideslang.validation import FidesKey
from loguru import logger

from fides.api.common_exceptions import ValidationError
from fides.api.graph.config import (
    Collection,
    CollectionAddress,
    EdgeDirection,
    FieldAddress,
    FieldPath,
    GraphDataset,
    SeedAddress,
)

DataCategoryFieldMapping = Dict[CollectionAddress, Dict[FidesKey, List[FieldPath]]]


class Node:
    """A traversal_node represents a single collection as a graph traversal_node.
    Note that a traversal_node is simply a store of a graph location and connections and does not imply ordering.

    An actual traversal is made up of Traversal nodes, which can be thought of as a reified instance of
    this graph encoding a specific traversal.

    Node children are any nodes that are reachable via this traversal_node.
    """

    def __init__(self, dataset: GraphDataset, collection: Collection):
        self.address = CollectionAddress(dataset.name, collection.name)
        self.dataset = dataset
        self.collection = collection

    def __repr__(self) -> str:
        return f"Node({self.address})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return False
        return other.address == self.address

    def __hash__(self) -> int:
        return hash(self.address)


class Edge:
    """A graph link uniquely defined by a pair of keys and a direction from f1->f2.
    Undirected edges are treated as a pair of f1->f2, f2->f2."""

    def __init__(self, f1: FieldAddress, f2: FieldAddress) -> None:
        if f1.collection_address() == f2.collection_address():
            raise ValidationError(f"collection self-reference not allowed {f1}<-->{f2}")
        self.f1 = f1
        self.f2 = f2

    def __repr__(self) -> str:
        return f"{self.f1}->{self.f2}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Edge):
            return False
        return other.f1 == self.f1 and other.f2 == self.f2

    def __hash__(self) -> int:
        return hash((self.f1, self.f2))

    def contains(self, node_address: CollectionAddress) -> bool:
        """The collection address is a prefix to either one of the endpoints."""
        return (
            self.f1.collection_address() == node_address
            or self.f2.collection_address() == node_address
        )

    def spans(self, addr_1: CollectionAddress, addr_2: CollectionAddress) -> bool:
        """True if the 2 provided addresses span the edge endpoints"""
        ds1 = self.f1.collection_address()
        ds2 = self.f2.collection_address()
        return ds1 == addr_1 and ds2 == addr_2

    def split_by_address(
        self, node_address: CollectionAddress
    ) -> Optional[Tuple[FieldAddress, FieldAddress]]:
        """Given the input traversal_node address, return the ends of this edge as an ordered pair (address, other) where
        the first element points to the input address and the second to the opposite side.
        """
        if self.f1.collection_address() == node_address:
            return self.f1, self.f2
        return None

    def ends_with_collection(self, addr: CollectionAddress) -> bool:
        """The far end of this edge points to the provided collection address"""
        return self.f2.collection_address() == addr

    @classmethod
    def delete_edges(
        cls,
        edges: Set[Edge],
        from_address: CollectionAddress,
        to_address: CollectionAddress,
    ) -> Set[Edge]:
        """Delete all edges between traversal_node address 1 and traversal_node address 2 and return the deleted edges."""
        to_delete = set(filter(lambda e: e.spans(from_address, to_address), edges))
        edges.difference_update(to_delete)
        return to_delete

    @classmethod
    def create_edge(
        cls,
        addr_1: FieldAddress,
        addr_2: FieldAddress,
        direction: Optional[EdgeDirection] = None,
    ) -> Edge:
        """Create an edge from addr_1 to addr_2 with the given direction."""
        if direction == "from":
            return Edge(addr_2, addr_1)
        if direction == "to":
            return Edge(addr_1, addr_2)
        return BidirectionalEdge(addr_1, addr_2)


class BidirectionalEdge(Edge):
    """A graph link whose direction is unspecified"""

    def __init__(self, f1: FieldAddress, f2: FieldAddress) -> None:
        super().__init__(f1, f2)
        if self.f2 < self.f1:
            self.f1, self.f2 = self.f2, self.f1

    def __repr__(self) -> str:
        return f"{self.f1}<->{self.f2}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BidirectionalEdge):
            return False
        return (other.f1 == self.f1 and other.f2 == self.f2) or (
            other.f1 == self.f2 and other.f2 == self.f1
        )

    def __hash__(self) -> int:
        return hash((self.f1, self.f2, True))

    def spans(self, addr_1: CollectionAddress, addr_2: CollectionAddress) -> bool:
        """True if the 2 provided addresses span the edge endpoints"""
        ds1 = self.f1.collection_address()
        ds2 = self.f2.collection_address()
        return (ds1 == addr_1 and ds2 == addr_2) or (ds1 == addr_2 and ds2 == addr_1)

    def split_by_address(
        self, node_address: CollectionAddress
    ) -> Optional[Tuple[FieldAddress, FieldAddress]]:
        """Given the input traversal_node address, return the ends of this edge as an ordered pair (address, other) where
        the first element points to the input address and the second to the opposite side.
        """
        if self.f1.collection_address() == node_address:
            return self.f1, self.f2
        if self.f2.collection_address() == node_address:
            return self.f2, self.f1
        return None

    def ends_with_collection(self, addr: CollectionAddress) -> bool:
        """The far end of this edge points to the provided collection address"""
        return self.contains(addr)


class DatasetGraph:
    """Graph representing the entirety of all addressable datasets.

    A graph holds a collection of nodes and an abstract graph. A reified traversal is
    generated only in reference to a specific seed, since the seeds determine which traversal_node
    (or nodes) represent the start nodes.
    """

    def __init__(self, *datasets: GraphDataset) -> None:
        """We create all edges based on field specifications.
        We also add child references to nodes. Note that this means that
        this is a destructive operation on the input datasets, as it
        will alter references within them"""

        # build nodes
        nodes = [Node(dr, ds) for dr in datasets for ds in dr.collections]
        self.nodes: dict[CollectionAddress, Node] = {
            node.address: node for node in nodes
        }

        # build links
        self.edges: Set[Edge] = set()

        for node_address, node in self.nodes.items():
            for field_path, ref_list in node.collection.references().items():
                source_field_address: FieldAddress = FieldAddress(
                    node_address.dataset, node_address.collection, *field_path.levels
                )
                for dest_field_address, direction in ref_list:
                    if dest_field_address.collection_address() not in self.nodes:
                        message = f"Referenced object {dest_field_address} from dataset {node_address.dataset} does not exist"
                        logger.warning(message)
                        raise ValidationError(
                            message, errors=[dest_field_address.value]
                        )
                    self.edges.add(
                        Edge.create_edge(
                            source_field_address, dest_field_address, direction
                        )
                    )

        # collect all seed references
        self.identity_keys: dict[FieldAddress, SeedAddress] = {
            FieldAddress(
                node.address.dataset,
                node.address.collection,
                *field_path.levels,
            ): seed_address
            for node in nodes
            for field_path, seed_address in node.collection.identities().items()
        }

    @property
    def data_category_field_mapping(
        self,
    ) -> DataCategoryFieldMapping:
        """
        Maps the data_categories for each traversal_node to a list of field paths that have that
        same data category.

        For example:
        {
            "postgres_example_test_dataset:address": {
                "user.contact.address.city": [FieldPath("city")],
                "user.contact.address.street": [FieldPath("house"), FieldPath("street")],
                "system.operations": [FieldPath("id")],
                "user.contact.address.state": [FieldPath("state"]),
                "user.contact.address.postal_code": [FieldPath("zip")]
            }
        }

        """
        mapping: Dict[CollectionAddress, Dict[FidesKey, List[FieldPath]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for node_address, node in self.nodes.items():
            mapping[node_address] = node.collection.field_paths_by_category
        return mapping

    def __repr__(self) -> str:
        return f"Graph: nodes = {self.nodes.keys()}"



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, Field, model_validator


class Operator(str, Enum):
    eq = "eq"
    neq = "neq"
    lt = "lt"
    lte = "lte"
    gt = "gt"
    gte = "gte"
    exists = "exists"
    not_exists = "not_exists"
    list_contains = "list_contains"
    not_in_list = "not_in_list"


class GroupOperator(str, Enum):
    and_ = "and"
    or_ = "or"


# Leaf condition (e.g., "user.name exists", "user.created_at >= 2024-01-01")
class ConditionLeaf(BaseModel):
    field_address: str = Field(
        description="Field path to check (e.g., 'user.name', 'billing.subscription.status')"
    )
    operator: Operator = Field(description="Operator to apply")
    value: Optional[
        Union[str, int, float, bool, List[Union[str, int, float, bool]]]
    ] = Field(default=None, description="Expected value for comparison")


# Nested logical condition (AND, OR)
class ConditionGroup(BaseModel):
    logical_operator: GroupOperator = Field(
        description="Logical operator: 'and' or 'or'"
    )
    conditions: list["Condition"] = Field(
        description="List of conditions or nested groups"
    )

    @model_validator(mode="after")
    def validate_conditions(self) -> "ConditionGroup":
        if not self.conditions:
            raise ValueError("conditions list cannot be empty")
        return self


# Use model_rebuild to allow recursive nesting
Condition = Union[ConditionLeaf, ConditionGroup]
ConditionGroup.model_rebuild()



"""index preference created and updated

Revision ID: 8342453518cc
Revises: 144d9b85c712
Create Date: 2023-04-19 18:33:17.034245

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8342453518cc"
down_revision = "144d9b85c712"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        op.f("ix_currentprivacypreference_created_at"),
        "currentprivacypreference",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_updated_at"),
        "currentprivacypreference",
        ["updated_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_created_at"),
        "privacypreferencehistory",
        ["created_at"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_privacypreferencehistory_created_at"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_updated_at"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_created_at"),
        table_name="currentprivacypreference",
    )



from rich_click import secho


class DockerCheckException(Exception):
    """An exception for use when checking & verifying Docker versions."""

    def __init__(self, message: str):
        self.message = message
        secho(f"ERROR: {self.message}", fg="red")
        raise SystemExit(1)



from fides.api.graph.config import CollectionAddress


class ManualTaskAddress:
    """Utility class for creating and parsing manual task addresses"""

    MANUAL_DATA_COLLECTION = "manual_data"

    @staticmethod
    def create(connection_config_key: str) -> CollectionAddress:
        """Create a CollectionAddress for manual data: {connection_key}:manual_data"""
        return CollectionAddress(
            dataset=connection_config_key,
            collection=ManualTaskAddress.MANUAL_DATA_COLLECTION,
        )

    @staticmethod
    def _is_manual_data_collection(collection_name: str) -> bool:
        """Check if collection name represents manual task data"""
        return collection_name == ManualTaskAddress.MANUAL_DATA_COLLECTION

    @staticmethod
    def is_manual_task_address(address: CollectionAddress) -> bool:
        """Check if address represents manual task data"""
        if isinstance(address, str):
            # Handle string format "connection_key:collection_name"
            _, _, collection_part = address.partition(":")
            if not collection_part:
                return False
            return ManualTaskAddress._is_manual_data_collection(collection_part)

        # Handle CollectionAddress object
        return ManualTaskAddress._is_manual_data_collection(address.collection)

    @staticmethod
    def get_connection_key(address: CollectionAddress) -> str:
        """Extract connection config key from manual task address"""
        if not ManualTaskAddress.is_manual_task_address(address):
            raise ValueError(f"Not a manual task address: {address}")

        if isinstance(address, str):
            # Handle string format "connection_key:collection_name"
            return address.split(":")[0]

        # Handle CollectionAddress object
        return address.dataset



"""make customfielddefinition uniqueness case insensitive

Revision ID: 15a3e7483249
Revises: e92da354691e
Create Date: 2023-05-02 15:03:56.256982

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.exc import IntegrityError

# revision identifiers, used by Alembic.
revision = "15a3e7483249"
down_revision = "e92da354691e"
branch_labels = None
depends_on = None


def upgrade():
    try:
        op.execute(
            """ CREATE UNIQUE INDEX ix_plus_custom_field_definition_unique_lowername_resourcetype
                ON plus_custom_field_definition
                (lower(name), resource_type)
            """
        )
    except IntegrityError as exc:
        raise Exception(
            f"Fides attempted to create new custom field definition unique index but got error: {exc}. "
            f"Adjust custom field names to avoid case-insensitive name overlaps for the same resource type."
        )

    # remove unnecessary index of unused field
    op.drop_index(
        "ix_plus_custom_field_definition_field_definition",
        table_name="plus_custom_field_definition",
    )


def downgrade():
    op.drop_index(
        op.f("ix_plus_custom_field_definition_unique_lowername_resourcetype"),
        table_name="plus_custom_field_definition",
    )

    # re-add unnecessray index of unused field in downgrade, for consistency
    op.create_index(
        op.f("ix_plus_custom_field_definition_field_definition"),
        "plus_custom_field_definition",
        ["field_definition"],
        unique=False,
    )



from typing import Any, Dict, List, Optional

from okta.client import Client as OktaClient
from okta.exceptions import OktaAPIException

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row
from fides.api.util.wrappers import sync


class OktaConnector(BaseConnector):
    """
    Okta connector for integrating with Okta's API.
    This connector allows for user management and authentication operations.
    """

    @property
    def dsr_supported(self) -> bool:
        return False

    def create_client(self) -> OktaClient:
        """Creates and returns an Okta client instance"""
        try:
            return OktaClient(
                {
                    "orgUrl": self.configuration.secrets["org_url"],
                    "token": self.configuration.secrets["api_token"],
                    "raiseException": True,
                }
            )
        except Exception as e:
            raise ConnectionException(f"Failed to create Okta client: {str(e)}")

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """Return the query config that corresponds to this connector type"""
        raise NotImplementedError("Query config not implemented for Okta")

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Validates the connection to Okta by attempting to list users.
        """
        try:
            self._list_applications()
            return ConnectionTestStatus.succeeded
        except OktaAPIException as e:
            error = e.args[0]
            raise ConnectionException(
                f"Failed to connect to Okta: {error['errorSummary']}"
            )
        except Exception as e:
            raise ConnectionException(
                f"Unexpected error testing Okta connection: {str(e)}"
            )

    @sync
    async def _list_applications(self) -> List[Dict[str, Any]]:
        """List all applications in Okta"""
        client = self.client()
        return await client.list_applications()

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """DSR execution not supported for Okta connector"""
        return []

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """DSR execution not supported for Okta connector"""
        return 0

    def close(self) -> None:
        """Close any held resources"""
        # No resources to close for Okta client



"""remove Shippo connection configs

Revision ID: fa476b733915
Revises: 59d9c4beeb30
Create Date: 2024-10-03 17:16:42.070361

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "fa476b733915"
down_revision = "59d9c4beeb30"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        WITH shippo_connections AS (
            SELECT id
            FROM connectionconfig
            WHERE connection_type = 'saas'
              AND saas_config->>'type' = 'shippo'
        )
        DELETE FROM datasetconfig
        WHERE connection_config_id IN (SELECT id FROM shippo_connections);

        DELETE FROM connectionconfig WHERE connection_type = 'saas' AND saas_config->>'type' = 'shippo';
    """
    )


def downgrade():
    # cannot downgrade since we can't restore a deleted connectionconfig
    pass



from typing import Dict

from celery.app.control import Inspect
from fastapi import APIRouter, HTTPException, Security
from starlette.status import HTTP_200_OK

from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.worker import QueueInfo, TaskDetails, WorkerInfo, WorkerStats
from fides.api.tasks import celery_app
from fides.api.util.cache import get_queue_counts
from fides.common.api.scope_registry import WORKER_STATS_READ
from fides.common.api.v1.urn_registry import V1_URL_PREFIX

router = APIRouter(tags=["Worker Stats"], prefix=V1_URL_PREFIX)


@router.get(
    "/worker-stats",
    dependencies=[Security(verify_oauth_client, scopes=[WORKER_STATS_READ])],
    status_code=HTTP_200_OK,
    response_model=WorkerStats,
)
def get_worker_stats() -> WorkerStats:  # pragma: no cover
    """Get statistics about Celery queues and workers.

    Returns information about task queues and worker states in the Celery system:
    - Queue statistics: Number of pending tasks in each Redis-backed queue
    - Worker statistics: Current state of each Celery worker including:
        - Active task: Currently executing task, if any
        - Reserved tasks: Tasks reserved by a worker but not yet started
        - Registered tasks: Tasks this worker is configured to accept
    """

    try:
        return WorkerStats(
            queues=get_queue_info(),
            workers=get_worker_info(),
        )
    except Exception as e:
        raise HTTPException(
            status_code=500, detail=f"Error getting worker stats: {str(e)}"
        )


def get_queue_info() -> Dict[str, QueueInfo]:
    """Return information about queue sizes"""
    return {
        queue_name: QueueInfo(messages=count)
        for queue_name, count in get_queue_counts().items()
    }


def get_worker_info() -> Dict[str, WorkerInfo]:
    inspector: Inspect = celery_app.control.inspect()
    active = inspector.active() or {}
    reserved = inspector.reserved() or {}
    registered = inspector.registered() or {}

    workers = {}

    # Process active tasks
    for worker_name, tasks in active.items():
        current_task = None
        if tasks and len(tasks) > 0:
            current_task = TaskDetails.from_celery_task(tasks[0], state="started")
        workers[worker_name] = WorkerInfo(active_task=current_task, reserved_tasks=[])

    # Process reserved tasks
    for worker_name, tasks in reserved.items():
        if worker_name not in workers:  # new worker we haven't seen
            workers[worker_name] = WorkerInfo(active_task=None, reserved_tasks=[])

        if tasks:
            reserved_tasks = [
                TaskDetails.from_celery_task(task, state="received") for task in tasks
            ]
            workers[worker_name].reserved_tasks = reserved_tasks

    for worker_name, registered_tasks in registered.items():
        workers[worker_name].registered_tasks = registered_tasks

    return workers



from typing import Optional

from fastapi import Depends
from sqlalchemy.orm import Session

from fides.api.api import deps
from fides.api.models.messaging import MessagingConfig
from fides.api.schemas.identity_verification import IdentityVerificationConfigResponse
from fides.api.util.api_router import APIRouter
from fides.common.api.v1 import urn_registry as urls
from fides.config.config_proxy import ConfigProxy

router = APIRouter(tags=["Identity Verification"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.ID_VERIFICATION_CONFIG,
    response_model=IdentityVerificationConfigResponse,
)
def get_id_verification_config(
    *,
    db: Session = Depends(deps.get_db),
    config_proxy: ConfigProxy = Depends(deps.get_config_proxy),
) -> IdentityVerificationConfigResponse:
    """Returns id verification config."""
    messaging_config: Optional[MessagingConfig] = db.query(MessagingConfig).first()
    return IdentityVerificationConfigResponse(
        identity_verification_required=config_proxy.execution.subject_identity_verification_required,
        disable_consent_identity_verification=config_proxy.execution.disable_consent_identity_verification,
        valid_email_config_exists=bool(messaging_config and messaging_config.secrets),
    )



from __future__ import annotations

from typing import List

from fastapi import HTTPException, status
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_404_NOT_FOUND,
)

from fides.common.api.scope_registry import SCOPE_REGISTRY as SCOPES


class FidesopsException(Exception):
    """Base class for fidesops exceptions"""

    def __init__(self, message: str, errors: List[str] = []):
        super().__init__(message)
        self.message = message
        self.errors = errors


class TraversalError(FidesopsException):
    """Error with the names of all nodes that could not be reached."""


class UnreachableNodesError(TraversalError):
    """Error with the names of all nodes that could not be reached, inherits from TraversalError."""


class UnreachableEdgesError(TraversalError):
    """Error with the names of all edges that could not be reached, inherits from TraversalError."""


class ValidationError(FidesopsException):
    """Data does not pass validation."""


class StorageUploadError(FidesopsException):
    """Data cannot be uploaded to storage destination"""


class SystemManagerException(FidesopsException):
    """Exception class when there are errors making a systemmanager"""


class ConnectionException(FidesopsException):
    """Exception class when there are errors making a connection"""


class InsufficientDataException(FidesopsException):
    """Exception class when there is not sufficient data to proceed"""


class SkippingConsentPropagation(BaseException):
    """Skipping consent propagation for collection. Used to trigger "skipped" execution logs being created where applicable
    for Privacy Preference requests on saas connectors.
    """


class RedisConnectionError(Exception):
    """The Configured Redis instance is uncontactable."""


class MisconfiguredPolicyException(Exception):
    """Thrown when a Privacy Request cannot be processed due to a misconfigured Policy."""


class PolicyValidationError(ValueError):
    """The policy you are trying to create has invalid data"""


class InvalidDataLengthValidationError(ValueError):
    """The length provided is invalid"""


class RuleValidationError(ValueError):
    """The Rule you are trying to create has invalid data"""


class StorageConfigValidationError(ValueError):
    """The Storage Config you are trying to create has invalid data"""


class InvalidDataTypeValidationError(ValueError):
    """The specified data type is invalid."""


class RuleTargetValidationError(ValueError):
    """The Rule you are trying to create has invalid data"""


class DataCategoryNotSupported(ValueError):
    """The data category you have supplied is not supported."""


class PolicyNotFoundException(Exception):
    """Policy could not be found"""


class ResumeTaskException(Exception):
    """Issue restoring data from collection to resume Privacy Request Processing"""


class ConnectorNotFoundException(Exception):
    """Connector could not be found"""


class DrpActionValidationError(Exception):
    """A resource already exists with this DRP Action."""


class StorageConfigNotFoundException(BaseException):
    """Custom Exception - StorageConfig Not Found"""


class SystemNotFound(BaseException):
    """System Not Found"""


class PrivacyNoticeHistoryNotFound(BaseException):
    """PrivacyNoticeHistory Not Found"""


class ConsentHistorySaveError(BaseException):
    """PrivacyPreferenceHistory or ServedNoticeHistory Save Error"""


class IdentityNotFoundException(BaseException):
    """Identity Not Found"""


class WebhookOrderException(BaseException):
    """Custom Exception - Issue with webhooks order"""


class PostProcessingException(BaseException):
    """Custom Exception - Issue with post processing"""


class CollectionDisabled(BaseException):
    """Collection is attached to disabled ConnectionConfig"""


class ActionDisabled(BaseException):
    """Collection is attached to a ConnectionConfig that has not enabled the given action"""


class NotSupportedForCollection(BaseException):
    """The given action is not supported for this type of collection"""


class PrivacyRequestExit(BaseException):
    """Privacy request exiting processing waiting on subtasks to complete"""


class PrivacyRequestCanceled(BaseException):
    """Privacy Request has been Canceled"""


class PrivacyRequestPaused(BaseException):
    """Halt Instruction Received on Privacy Request"""


class PrivacyRequestNotFound(BaseException):
    """Privacy Request Not Found"""


class RequestTaskNotFound(BaseException):
    """Privacy Request Task Not Found"""


class AwaitingAsyncTaskCallback(BaseException):
    """Request Task is Awaiting Processing - Awaiting Async Task Callback"""


class UpstreamTasksNotReady(BaseException):
    """Privacy Request Task awaiting upstream tasks"""


class NoCachedManualWebhookEntry(BaseException):
    """No manual data exists for this webhook on the given privacy request."""


class ManualWebhookFieldsUnset(BaseException):
    """Manual webhook has fields that are not explicitly set: Likely new field has been added"""


class PrivacyRequestErasureEmailSendRequired(BaseException):
    """Erasure requests will need to be fulfilled by email send.  Exception is raised to change ExecutionLog details"""


class SaaSConfigNotFoundException(FidesopsException):
    """Custom Exception - SaaS Config Not Found"""


class MessagingConfigNotFoundException(FidesopsException):
    """Custom Exception - Messaging Config Not Found"""


class MessagingConfigValidationException(FidesopsException):
    """Custom Exception - Messaging Config Could Not Be Created, Updated, or Deleted"""


class MessageDispatchException(FidesopsException):
    """Custom Exception - Message Dispatch Error"""


class EmailTemplateUnhandledActionType(FidesopsException):
    """Custom Exception - Email Template Unhandled ActionType Error"""


class EmailTemplateNotFoundException(FidesopsException):
    """Custom Exception - Email Template Not Found"""


class MessagingTemplateValidationException(FidesopsException):
    """Custom Exception - Messaging Template Could Not Be Created, Updated, or Deleted"""


class OAuth2TokenException(FidesopsException):
    """Custom Exception - Unable to access or refresh OAuth2 tokens for SaaS connector"""


class AuthenticationFailure(HTTPException):
    """Wrapper for authentication failure exception"""

    def __init__(self, detail: str) -> None:
        super().__init__(status_code=HTTP_401_UNAUTHORIZED, detail=detail)


class BadRequest(HTTPException):
    """Wrapper for bad request exception"""

    def __init__(self, detail: str) -> None:
        super().__init__(status_code=HTTP_400_BAD_REQUEST, detail=detail)


class NotFoundException(HTTPException):
    """Wrapper for not found exception"""

    def __init__(self, detail: str) -> None:
        super().__init__(status_code=HTTP_404_NOT_FOUND, detail=detail)


class ClientUnsuccessfulException(FidesopsException):
    """Exception for when client call fails"""

    def __init__(self, status_code: int):
        super().__init__(message=f"Client call failed with status code '{status_code}'")


class NoSuchStrategyException(ValueError):
    """Exception for when a masking strategy does not exist"""


class RedisNotConfigured(Exception):
    """Redis cache is unavailable due to configuration"""


class InvalidSaaSRequestOverrideException(ValueError):
    """Exception for when a provied SaaS request override function is invalid"""


class NoSuchSaaSRequestOverrideException(ValueError):
    """Exception for when a requested SaaS request override function does not exist"""


class IdentityVerificationException(FidesopsException):
    """Custom exceptions for when we cannot verify the identity of a subjct"""


class NoSuchConnectionTypeSecretSchemaError(Exception):
    """Exception for when a connection type secret schema is not found."""


class SSHTunnelConfigNotFoundException(Exception):
    """Exception for when Fides is configured to use an SSH tunnel without config provided."""


class MalisciousUrlException(Exception):
    """Fides has detected a potentially maliscious URL."""


class MaskingSecretsExpired(BaseException):
    """The cached masking secrets have expired for the given privacy request."""


class AuthenticationError(HTTPException):
    """To be raised when attempting to fetch an access token using
    invalid credentials.
    """

    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
        )


class AuthorizationError(HTTPException):
    """Throws an HTTP 403"""

    def __init__(self, detail: str) -> None:
        """Override the regular HTTPException throwing only a 403"""
        super().__init__(status_code=status.HTTP_403_FORBIDDEN, detail=detail)


class ClientWriteFailedError(HTTPException):
    """To be raised when a client cannot be created."""

    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Failed to create client",
        )


class ClientNotFoundError(HTTPException):
    """To be raised when attempting to fetch a client that does not exist."""

    def __init__(self, client_id: str) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail={
                "error": "Client does not exist",
                "id": client_id,
            },
        )


class ExpiredTokenError(HTTPException):
    """To be raised when a provided token is expired."""

    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="OAuth token expired",
        )


class InvalidAuthorizationSchemeError(HTTPException):
    """To be raised when attempting to authenticate with an unexpected
    Authorization header value.
    """

    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Failed to authenticate",
            headers={"WWW-Authenticate": "Bearer"},
        )


class InvalidScopeError(HTTPException):
    """To be raised when a provided scope does not exist."""

    def __init__(self, invalid_scopes: list[str]) -> None:
        SCOPES.sort()

        super().__init__(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail={
                "error": "Invalid scope provided",
                "invalid_scopes": invalid_scopes,
                "valid_scopes": SCOPES,
            },
        )


class KeyOrNameAlreadyExists(Exception):
    """A resource already exists with this key or name."""


class KeyValidationError(Exception):
    """The resource you're trying to create has a key specified but not
    a name specified.
    """


class MissingConfig(Exception):
    """Custom exception for when no valid configuration file is provided."""


class MonitorConfigNotFoundException(BaseException):
    """MonitorConfig could not be found"""


class MissingNamespaceSchemaException(BaseException):
    """Attempting to use namespace fides_meta without specifying the schema to validate it."""



from typing import List

from loguru import logger
from sqlalchemy_bigquery import BigQueryDialect

from fides.api.schemas.partitioning.time_based_partitioning import TimeBasedPartitioning


class BigQueryTimeBasedPartitioning(TimeBasedPartitioning):
    """Generates BigQuery-specific WHERE clauses for time-based partitioning."""

    def generate_where_clauses(self) -> List[str]:
        """Generate BigQuery-specific WHERE clauses."""
        conditions = self.generate_expressions()
        bigquery_dialect = BigQueryDialect()

        partition_clauses = []
        for condition in conditions:
            try:
                clause_str = str(
                    condition.compile(
                        dialect=bigquery_dialect,
                        compile_kwargs={"literal_binds": True},
                    )
                )
                partition_clauses.append(clause_str)
            except Exception as exc:
                logger.error("Could not generate where clause", exc_info=True)
                raise exc

        return partition_clauses



"""Adds system managers table and client.systems

Also includes a data migration to add new scopes to fidesuserpermission.scopes and client.scopes if they
have cli-objects:-* scopes as we're making these scopes more granular.

Revision ID: 68a518a3c050
Revises: 2f6aa5fe797a
Create Date: 2023-02-23 21:52:56.225405

"""

from typing import List

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import ResultProxy
from sqlalchemy.sql.elements import TextClause

revision = "68a518a3c050"
down_revision = "2f6aa5fe797a"
branch_labels = None
depends_on = None

new_create_scopes = [
    "data_category:create",
    "data_qualifier:create",
    "data_subject:create",
    "data_use:create",
    "ctl_dataset:create",
    "evaluation:create",
    "organization:create",
    "ctl_policy:create",
    "registry:create",
    "system:create",
]

new_read_scopes = [
    "data_category:read",
    "data_qualifier:read",
    "data_subject:read",
    "data_use:read",
    "ctl_dataset:read",
    "evaluation:read",
    "organization:read",
    "ctl_policy:read",
    "registry:read",
    "system:read",
]

new_update_scopes = [
    "data_category:update",
    "data_qualifier:update",
    "data_subject:update",
    "data_use:update",
    "ctl_dataset:update",
    "evaluation:update",
    "organization:update",
    "ctl_policy:update",
    "registry:update",
    "system:update",
]

new_delete_scopes = [
    "data_category:delete",
    "data_qualifier:delete",
    "data_subject:delete",
    "data_use:delete",
    "ctl_dataset:delete",
    "evaluation:delete",
    "organization:delete",
    "ctl_policy:delete",
    "registry:delete",
    "system:delete",
]


def upgrade():
    op.create_table(
        "systemmanager",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("user_id", sa.String(), nullable=False),
        sa.Column("system_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(["system_id"], ["ctl_systems.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fidesuser.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id", "user_id", "system_id"),
    )
    op.create_index(op.f("ix_systemmanager_id"), "systemmanager", ["id"], unique=False)
    op.add_column(
        "client",
        sa.Column(
            "systems", sa.ARRAY(sa.String()), server_default="{}", nullable=False
        ),
    )

    bind = op.get_bind()

    def add_new_scopes_to_fides_user_permissions(
        matched_rows: ResultProxy, new_scopes: List[str]
    ):
        """Helper method to tack on new scopes to the fidesuserpermissions.scopes if user has existing scope."""
        for row in matched_rows:
            scopes: List[str] = row["scopes"] or []
            scopes.extend(new_scopes)

            add_new_scopes_query: TextClause = text(
                "UPDATE fidesuserpermissions SET scopes= :scopes WHERE id= :id"
            )

            bind.execute(
                add_new_scopes_query,
                {"scopes": sorted(list(set(scopes))), "id": row["id"]},
            )

    def add_new_scopes_to_client(matched_rows: ResultProxy, new_scopes: List[str]):
        """Helper method to tack on new scopes to the client.scopes if user has existing scope."""
        for row in matched_rows:
            scopes: List[str] = row["scopes"] or []
            scopes.extend(new_scopes)

            add_new_scopes_query: TextClause = text(
                "UPDATE client SET scopes= :scopes WHERE id= :id"
            )

            bind.execute(
                add_new_scopes_query,
                {"scopes": sorted(list(set(scopes))), "id": row["id"]},
            )

    add_new_scopes_to_fides_user_permissions(
        matched_rows=bind.execute(
            text(
                "SELECT * FROM fidesuserpermissions WHERE 'cli-objects:create' = ANY(scopes);"
            )
        ),
        new_scopes=new_create_scopes,
    )
    add_new_scopes_to_client(
        matched_rows=bind.execute(
            text("SELECT * FROM client WHERE 'cli-objects:create' = ANY(scopes);")
        ),
        new_scopes=new_create_scopes,
    )

    add_new_scopes_to_fides_user_permissions(
        matched_rows=bind.execute(
            text(
                "SELECT * FROM fidesuserpermissions WHERE 'cli-objects:read' = ANY(scopes);"
            )
        ),
        new_scopes=new_read_scopes,
    )
    add_new_scopes_to_client(
        matched_rows=bind.execute(
            text("SELECT * FROM client WHERE 'cli-objects:read' = ANY(scopes);")
        ),
        new_scopes=new_read_scopes,
    )

    add_new_scopes_to_fides_user_permissions(
        matched_rows=bind.execute(
            text(
                "SELECT * FROM fidesuserpermissions WHERE 'cli-objects:update' = ANY(scopes);"
            )
        ),
        new_scopes=new_update_scopes,
    )

    add_new_scopes_to_client(
        matched_rows=bind.execute(
            text("SELECT * FROM client WHERE 'cli-objects:update' = ANY(scopes);")
        ),
        new_scopes=new_update_scopes,
    )

    add_new_scopes_to_fides_user_permissions(
        matched_rows=bind.execute(
            text(
                "SELECT * FROM fidesuserpermissions WHERE 'cli-objects:delete' = ANY(scopes);"
            )
        ),
        new_scopes=new_delete_scopes,
    )

    add_new_scopes_to_client(
        matched_rows=bind.execute(
            text("SELECT * FROM client WHERE 'cli-objects:delete' = ANY(scopes);")
        ),
        new_scopes=new_delete_scopes,
    )


def downgrade():
    op.drop_column("client", "systems")
    op.drop_index(op.f("ix_systemmanager_id"), table_name="systemmanager")
    op.drop_table("systemmanager")



from enum import Enum
from typing import List, Optional

from pydantic import ConfigDict, EmailStr, field_validator

from fides.api.custom_types import PhoneNumber
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.policy import DrpAction

DRP_VERSION = "0.5"


class DrpMeta(FidesSchema):
    """Enum to hold Drp metadata. Only version is supported at this time"""

    version: str


class DrpRegime(Enum):
    """Enum to hold Drp Regime. Only ccpa supported at this time"""

    ccpa = "ccpa"


class DrpPrivacyRequestCreate(FidesSchema):
    """Data required to create a DRP PrivacyRequest"""

    meta: DrpMeta
    regime: Optional[DrpRegime] = None
    exercise: List[DrpAction]
    relationships: Optional[List[str]] = None
    identity: str
    status_callback: Optional[str] = None
    model_config = ConfigDict(use_enum_values=True)

    @field_validator("exercise")
    @classmethod
    def check_exercise_length(cls, exercise: List[DrpAction]) -> List[DrpAction]:
        """Validate the only one exercise action is provided"""
        if len(exercise) > 1:
            raise ValueError("Multiple exercise actions are not supported at this time")
        return exercise


class DrpIdentity(FidesSchema):
    """Drp identity props"""

    aud: Optional[str] = None
    sub: Optional[str] = None
    name: Optional[str] = None
    email: Optional[EmailStr] = None
    email_verified: Optional[bool] = None
    phone_number: Optional[PhoneNumber] = None
    phone_number_verified: Optional[bool] = None
    address: Optional[str] = None
    address_verified: Optional[bool] = None
    owner_of_attorney: Optional[str] = None


class DrpDataRightsResponse(FidesSchema):
    """Drp data rights response"""

    version: str
    api_base: Optional[str] = None
    actions: List[DrpAction]
    user_relationships: Optional[List[str]] = None


class DrpRevokeRequest(FidesSchema):
    """DRP Data Rights Revoke Request Body"""

    request_id: str
    reason: Optional[str] = None



from __future__ import annotations

import json
from typing import Any, Callable, Dict, List, Optional, Union

import requests
import sendgrid
from loguru import logger
from sendgrid.helpers.mail import Content, Email, Mail, Personalization, TemplateId, To
from sqlalchemy.orm import Session
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from fides.api.common_exceptions import MessageDispatchException
from fides.api.email_templates.get_email_template import get_email_template
from fides.api.models.messaging import (  # type: ignore[attr-defined]
    EMAIL_MESSAGING_SERVICES,
    MessagingConfig,
    get_messaging_method,
)
from fides.api.models.messaging_template import MessagingTemplate
from fides.api.schemas.messaging.messaging import (
    CONFIGURABLE_MESSAGING_ACTION_TYPES,
    AccessRequestCompleteBodyParams,
    ConsentEmailFulfillmentBodyParams,
    EmailForActionType,
    ErasureRequestBodyParams,
    ErrorNotificationBodyParams,
    ExternalUserWelcomeBodyParams,
    FidesopsMessage,
    MessagingActionType,
    MessagingMethod,
    MessagingServiceDetails,
    MessagingServiceSecrets,
    MessagingServiceType,
    RequestReceiptBodyParams,
    RequestReviewDenyBodyParams,
    SubjectIdentityVerificationBodyParams,
    UserInviteBodyParams,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.messaging.messaging_crud_service import (
    get_basic_messaging_template_by_type_or_default,
    get_enabled_messaging_template_by_type_and_property,
)
from fides.api.tasks import DatabaseTask, celery_app
from fides.api.util.logger import Pii
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy
from fides.service.messaging.aws_ses_service import AWS_SES_Service

EMAIL_JOIN_STRING = ", "
EMAIL_TEMPLATE_NAME = "fides"


@celery_app.task(base=DatabaseTask, bind=True)
def dispatch_message_task(
    self: DatabaseTask,
    message_meta: Dict[str, Any],
    service_type: Optional[str],
    to_identity: Dict[str, Any],
    property_id: Optional[str],
) -> None:
    """
    A wrapper function to dispatch a message task into the Celery queues
    """
    schema = FidesopsMessage.model_validate(message_meta)
    with self.get_new_session() as db:
        dispatch_message(
            db=db,
            action_type=schema.action_type,
            to_identity=Identity.model_validate(to_identity),
            service_type=service_type,
            message_body_params=schema.body_params,
            property_id=property_id,
        )


def _property_specific_messaging_eligible(db: Session) -> bool:
    """
    Helper method to determine whether property specific messaging is eligible at all. To be eligible:
    1. The ENV must be configured
    2. The messaging type must be set as email. SMS is not yet supported for property-specific messaging.

    This method does not include a check for valid and enabled templates.
    """
    property_specific_messaging_enabled = ConfigProxy(
        db
    ).notifications.enable_property_specific_messaging
    if not property_specific_messaging_enabled:
        return False

    # Only email messaging method is supported when property-specific messaging is enabled.
    service_type = get_email_messaging_config_service_type(db=db)
    if not service_type or get_messaging_method(service_type) != MessagingMethod.EMAIL:
        logger.warning(
            "An email messaging config must be configured if property specific messaging is enabled."
        )
        return False
    return True


def get_property_specific_messaging_template(
    db: Session, property_id: Optional[str], action_type: MessagingActionType
) -> Optional[MessagingTemplate]:
    """
    Returns specific messaging template if one is enabled for a given action type and property.
    """
    is_property_specific_messaging_eligible = _property_specific_messaging_eligible(db)
    if not is_property_specific_messaging_eligible:
        return None

    template = get_enabled_messaging_template_by_type_and_property(
        db=db,
        template_type=action_type.value,
        property_id=property_id,
        use_default_property=True,
    )
    if not template:
        return None
    return template


def message_send_enabled(
    db: Session,
    property_id: Optional[str],
    action_type: MessagingActionType,
    basic_email_template_enabled: bool,
) -> bool:
    """
    Determines whether sending messages from Fides is enabled or disabled.

    Assumes action_type is one of CONFIGURABLE_MESSAGING_ACTION_TYPES.

    Property-specific messaging, if enabled, always takes precedence, and requires checking "enabled" templates for the
    given action type and property.
    """
    property_specific_messaging_enabled = ConfigProxy(
        db
    ).notifications.enable_property_specific_messaging

    if property_specific_messaging_enabled:
        # Only email messaging method is supported when property-specific messaging is enabled.
        service_type = get_email_messaging_config_service_type(db=db)
        if (
            not service_type
            or get_messaging_method(service_type) != MessagingMethod.EMAIL
        ):
            logger.warning(
                "An email messaging config must be configured if property specific messaging is enabled."
            )
            return False
        property_specific_messaging_template = get_property_specific_messaging_template(
            db=db, property_id=property_id, action_type=action_type
        )
        if property_specific_messaging_template:
            return True
    elif basic_email_template_enabled:
        return True
    logger.info("Message send is disabled for action type {}", action_type)
    return False


def dispatch_message(
    db: Session,
    action_type: MessagingActionType,
    *,
    to_identity: Optional[Identity],
    service_type: Optional[str],
    message_body_params: Optional[
        Union[
            AccessRequestCompleteBodyParams,
            ConsentEmailFulfillmentBodyParams,
            SubjectIdentityVerificationBodyParams,
            RequestReceiptBodyParams,
            RequestReviewDenyBodyParams,
            ErasureRequestBodyParams,
            UserInviteBodyParams,
            ErrorNotificationBodyParams,
            ExternalUserWelcomeBodyParams,
        ]
    ] = None,
    subject_override: Optional[str] = None,
    property_id: Optional[str] = None,
) -> None:
    """
    Sends a message to `to_identity` with content supplied in `message_body_params`
    """
    if not to_identity:
        logger.error("Message failed to send. No identity supplied.")
        raise MessageDispatchException("No identity supplied.")
    if not service_type:
        logger.error("Message failed to send. No notification service type configured.")
        raise MessageDispatchException("No notification service type configured.")

    logger.info("Retrieving message config")
    messaging_config: MessagingConfig = MessagingConfig.get_configuration(
        db=db, service_type=service_type
    )
    logger.info(
        "Building appropriate message template for action type: {}", action_type
    )
    messaging_method = get_messaging_method(service_type)
    message: Optional[Union[EmailForActionType, str]] = None
    messaging_template: Optional[MessagingTemplate] = None

    # If property-specific messaging is enabled and message type is one of the configurable templates,
    # we switch over to this mode, regardless of other ENV vars
    if (
        ConfigProxy(db).notifications.enable_property_specific_messaging
        and action_type in CONFIGURABLE_MESSAGING_ACTION_TYPES
    ):
        property_specific_messaging_template = get_property_specific_messaging_template(
            db, property_id, action_type
        )
        if not property_specific_messaging_template:
            logger.warning(
                "Skipping sending property-specific email as no enabled template was found for action type: {}",
                action_type,
            )
            return
        messaging_template = property_specific_messaging_template
    else:
        logger.info(
            "Getting custom messaging template for action type: {}", action_type
        )
        messaging_template = get_basic_messaging_template_by_type_or_default(
            db=db, template_type=action_type.value
        )

    config_proxy = ConfigProxy(db=db)

    if messaging_method == MessagingMethod.EMAIL:
        message = _build_email(
            config_proxy=config_proxy,
            action_type=action_type,
            body_params=message_body_params,
            messaging_template=messaging_template,
        )
    elif messaging_method == MessagingMethod.SMS:
        message = _build_sms(
            action_type=action_type,
            body_params=message_body_params,
        )
    else:  # pragma: no cover
        # This is here as a fail safe, but it should be impossible to reach because
        # is controlled by a database enum field.
        logger.error("Notification service type is not valid: {}", service_type)
        raise MessageDispatchException(
            f"Notification service type is not valid: {service_type}"
        )
    messaging_service: MessagingServiceType = messaging_config.service_type  # type: ignore
    logger.info(
        "Retrieving appropriate dispatcher for email service: {}", messaging_service
    )
    dispatcher: Optional[Callable] = _get_dispatcher_from_config_type(
        message_service_type=messaging_service
    )
    if not dispatcher:
        logger.error(
            "Dispatcher has not been implemented for message service type: {}",
            messaging_service,
        )
        raise MessageDispatchException(
            f"Dispatcher has not been implemented for message service type: {messaging_service}"
        )
    logger.info(
        "Starting message dispatch for messaging service with action type: {}",
        action_type,
    )

    if subject_override and isinstance(message, EmailForActionType):
        message.subject = subject_override

    to = (
        to_identity.email
        if messaging_method == MessagingMethod.EMAIL
        else to_identity.phone_number
    )

    if not to:
        error_message = f"No {'email' if messaging_method == MessagingMethod.EMAIL else 'phone'} identity supplied."
        logger.error(f"Message failed to send. {error_message}")
        raise MessageDispatchException(error_message)

    dispatcher(
        messaging_config,
        message,
        to,
    )


def _build_sms(  # pylint: disable=too-many-return-statements
    action_type: MessagingActionType,
    body_params: Any,
) -> str:
    separator = ","
    if action_type == MessagingActionType.SUBJECT_IDENTITY_VERIFICATION:
        return (
            f"Your privacy request verification code is {body_params.verification_code}. "
            f"Please return to the Privacy Center and enter the code to continue. "
            f"This code will expire in {body_params.get_verification_code_ttl_minutes()} minutes"
        )
    if action_type == MessagingActionType.CONSENT_REQUEST:
        return (
            f"Your consent request verification code is {body_params.verification_code}. "
            "Please return to the consent request page and enter the code to continue. "
            f"This code will expire in {body_params.get_verification_code_ttl_minutes()} minutes"
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_RECEIPT:
        if len(body_params.request_types) > 1:
            return f"The following requests have been received: {separator.join(body_params.request_types)}"
        return f"Your {body_params.request_types[0]} request has been received"
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS:
        # Converting the expiration time to days
        subject_request_download_time_in_days = (
            CONFIG.security.subject_request_download_link_ttl_seconds / 86400
        )
        if len(body_params.download_links) > 1:
            return (
                "Your data access has been completed and can be downloaded at the following links. "
                f"For security purposes, these secret links will expire in {subject_request_download_time_in_days} days: "
                f"{separator.join(body_params.download_links)}"
            )
        return (
            f"Your data access has been completed and can be downloaded at {body_params.download_links[0]}. "
            f"For security purposes, this secret link will expire in {subject_request_download_time_in_days} days."
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION:
        return "Your privacy request for deletion has been completed."
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE:
        return "Your privacy request has been approved and is currently processing."
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY:
        if body_params.rejection_reason:
            return f"Your privacy request has been denied for the following reason: {body_params.rejection_reason}"
        return "Your privacy request has been denied."
    if action_type == MessagingActionType.TEST_MESSAGE:
        return "Test message from Fides."
    logger.error("Message action type {} is not implemented", action_type)
    raise MessageDispatchException(
        f"Message action type {action_type} is not implemented"
    )


def _render(template_str: str, variables: Optional[Dict] = None) -> str:
    """Helper function to render a template string with the provided variables."""
    if variables is None:
        variables = {}

    for key, value in variables.items():
        template_str = template_str.replace(f"__{key.upper()}__", str(value))

    return template_str


def _build_email(  # pylint: disable=too-many-return-statements, too-many-branches
    config_proxy: ConfigProxy,
    action_type: MessagingActionType,
    body_params: Any,
    messaging_template: Optional[MessagingTemplate] = None,
) -> EmailForActionType:
    """
    Builds an email for a specified messaging action type, using the provided parameters.

    The messaging_template parameter is used as the template for the email wording, and
    its rendered output is passed to the HTML templates. This is only applicable for action
    types that allow the user to specify a custom messaging template.

    Parameters:
        action_type (MessagingActionType): The type of messaging action for which the email is being built.
        body_params (Any): Parameters used to populate the email body, such as verification codes.
        messaging_template (Optional[MessagingTemplate]): An optional custom messaging template for the email wording.
            This parameter is used to define the subject and body of the email, and its rendered output is
            passed to the HTML templates. Only applicable for action types in CONFIGURABLE_MESSAGING_ACTION_TYPES.

    Returns:
        EmailForActionType: The constructed email object with the subject and body populated based on the action type.
    """
    if action_type == MessagingActionType.SUBJECT_IDENTITY_VERIFICATION:
        variables = {
            "code": body_params.verification_code,
            "minutes": body_params.get_verification_code_ttl_minutes(),
        }
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"], variables),  # type: ignore
            body=_render(messaging_template.content["body"], variables),  # type: ignore
            template_variables=variables,
        )
    if action_type == MessagingActionType.MESSAGE_ERASURE_REQUEST_FULFILLMENT:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject=f"Notification of user erasure requests from {body_params.controller}",
            body=base_template.render(
                {
                    "controller": body_params.controller,
                    "third_party_vendor_name": body_params.third_party_vendor_name,
                    "identities": body_params.identities,
                }
            ),
        )
    if action_type == MessagingActionType.CONSENT_REQUEST_EMAIL_FULFILLMENT:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Notification of users' consent preference changes",
            body=base_template.render({"body": body_params}),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_RECEIPT:
        variables = {"request_types": body_params.request_types}
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"], variables),  # type: ignore
            body=_render(messaging_template.content["body"], variables),  # type: ignore
            template_variables=variables,
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS:
        variables = {
            "download_link": body_params.download_links[0],
            "days": body_params.subject_request_download_time_in_days,
        }
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"], variables),  # type: ignore
            body=_render(messaging_template.content["body"], variables),  # type: ignore
            template_variables=variables,
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION:
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"]),  # type: ignore
            body=_render(messaging_template.content["body"]),  # type: ignore
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_ERROR_NOTIFICATION:
        # This action type does not use the default templates that are configurable in the Admin-UI.
        # They are instead hard-coded in fides, which we retrieve using get_email_template(action_type)
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Privacy Request Error Alert",
            body=base_template.render(),
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE:
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"]),  # type: ignore
            body=_render(messaging_template.content["body"]),  # type: ignore
        )
    if action_type == MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY:
        variables = {"denial_reason": body_params.rejection_reason}
        return EmailForActionType(
            subject=_render(messaging_template.content["subject"], variables),  # type: ignore
            body=_render(messaging_template.content["body"], variables),  # type: ignore
            template_variables=variables,
        )
    if action_type == MessagingActionType.TEST_MESSAGE:
        # This action type does not use the default templates that are configurable in the Admin-UI.
        # They are instead hard-coded in fides, which we retrieve using get_email_template(action_type)
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Test message from fides", body=base_template.render()
        )
    if action_type == MessagingActionType.USER_INVITE:
        base_template = get_email_template(action_type)
        return EmailForActionType(
            subject="Welcome to Fides",
            body=base_template.render(
                {
                    "admin_ui_url": config_proxy.admin_ui.url,
                    "username": body_params.username,
                    "invite_code": body_params.invite_code,
                }
            ),
        )
    if action_type == MessagingActionType.EXTERNAL_USER_WELCOME:
        base_template = get_email_template(action_type)
        # Generate display name for personalization
        display_name = body_params.username
        if body_params.first_name:
            display_name = body_params.first_name
            if body_params.last_name:
                display_name = f"{body_params.first_name} {body_params.last_name}"

        portal_link = (
            f"{body_params.privacy_center_url}?access_token={body_params.access_token}"
        )

        variables = {
            "username": body_params.username,
            "display_name": display_name,
            "first_name": body_params.first_name,
            "last_name": body_params.last_name,
            "org_name": body_params.org_name,
            "portal_link": portal_link,
            "privacy_center_url": body_params.privacy_center_url,
            "access_token": body_params.access_token,
        }

        return EmailForActionType(
            subject="Welcome to our Privacy Center",
            body=base_template.render(variables),
            template_variables=variables,
        )

    logger.error("Message action type {} is not implemented", action_type)
    raise MessageDispatchException(
        f"Message action type {action_type} is not implemented"
    )


def _get_dispatcher_from_config_type(
    message_service_type: MessagingServiceType,
) -> Optional[Callable]:
    """Determines which dispatcher to use based on message service type"""
    handler = {
        MessagingServiceType.mailgun: _mailgun_dispatcher,
        MessagingServiceType.mailchimp_transactional: _mailchimp_transactional_dispatcher,
        MessagingServiceType.twilio_text: _twilio_sms_dispatcher,
        MessagingServiceType.twilio_email: _twilio_email_dispatcher,
        MessagingServiceType.aws_ses: _aws_ses_dispatcher,
    }
    return handler.get(message_service_type)  # type: ignore


def validate_config(
    messaging_config: MessagingConfig,
    config_name: str,
    validate_details: Optional[bool] = True,
) -> None:
    """
    Validates that the messaging config has the required details and secrets.
    """
    condition = (
        not messaging_config.details or not messaging_config.secrets
        if validate_details
        else not messaging_config.secrets
    )

    if condition:
        error_message = f"No {config_name} config {'details or secrets' if validate_details else 'secrets'} supplied."
        logger.error(f"Message failed to send. {error_message}")
        raise MessageDispatchException(error_message)


def _mailchimp_transactional_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: str,
) -> None:
    """Dispatches email using Mailchimp Transactional"""
    validate_config(messaging_config, "Mailchimp Transactional")

    from_email = messaging_config.details[MessagingServiceDetails.EMAIL_FROM.value]
    data = json.dumps(
        {
            "key": messaging_config.secrets[
                MessagingServiceSecrets.MAILCHIMP_TRANSACTIONAL_API_KEY.value
            ],
            "message": {
                "from_email": from_email,
                "subject": message.subject,
                "html": message.body,
                # On Mailchimp Transactional's free plan `to` must be an email of the same
                # domain as `from_email`
                "to": [{"email": to.strip(), "type": "to"}],
            },
        }
    )

    response = requests.post(
        "https://mandrillapp.com/api/1.0/messages/send",
        headers={"Content-Type": "application/json"},
        data=data,
    )
    if not response.ok:
        logger.error("Email failed to send with status code: %s", response.status_code)
        raise MessageDispatchException(
            f"Email failed to send with status code {response.status_code}"
        )

    send_data = response.json()[0]
    email_rejected = send_data.get("status", "rejected") == "rejected"
    if email_rejected:
        reason = send_data.get("reject_reason", "Fides Error")
        explanations = {
            "soft-bounce": "A temporary error occured with the target inbox. For example, this inbox could be full. See https://mailchimp.com/developer/transactional/docs/reputation-rejections/#bounces for more info.",
            "hard-bounce": "A permanent error occured with the target inbox. See https://mailchimp.com/developer/transactional/docs/reputation-rejections/#bounces for more info.",
            "recipient-domain-mismatch": f"You are not authorised to send email to this domain from {from_email}.",
            "unsigned": f"The sending domain for {from_email} has not been fully configured for Mailchimp Transactional. See https://mailchimp.com/developer/transactional/docs/authentication-delivery/#authentication/ for more info.",
        }
        explanation = explanations.get(reason, "")
        raise MessageDispatchException(
            f"Verification email unable to send due to reason: {reason}. {explanation}"
        )


def _mailgun_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: str,
) -> None:
    """Dispatches email using Mailgun"""
    validate_config(messaging_config, "Mailgun")

    base_url = (
        "https://api.mailgun.net"
        if messaging_config.details[MessagingServiceDetails.IS_EU_DOMAIN.value] is False
        else "https://api.eu.mailgun.net"
    )

    domain = messaging_config.details[MessagingServiceDetails.DOMAIN.value]

    try:
        # Check if a fides template exists
        template_test = requests.get(
            f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/templates/{EMAIL_TEMPLATE_NAME}",
            auth=(
                "api",
                messaging_config.secrets[MessagingServiceSecrets.MAILGUN_API_KEY.value],
            ),
        )

        data = {
            "from": f"<mailgun@{domain}>",
            "to": [to.strip()],
            "subject": message.subject,
        }

        if template_test.status_code == 200:
            mailgun_variables = {
                "fides_email_body": message.body,
                **(message.template_variables or {}),
            }
            data["template"] = EMAIL_TEMPLATE_NAME
            data["h:X-Mailgun-Variables"] = json.dumps(mailgun_variables)
            response = requests.post(
                f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/messages",
                auth=(
                    "api",
                    messaging_config.secrets[
                        MessagingServiceSecrets.MAILGUN_API_KEY.value
                    ],
                ),
                data=data,
            )

            if not response.ok:
                logger.error(
                    "Email failed to send with status code: %s", response.status_code
                )
                raise MessageDispatchException(
                    f"Email failed to send with status code {response.status_code}"
                )
        else:
            data["html"] = message.body
            response = requests.post(
                f"{base_url}/{messaging_config.details[MessagingServiceDetails.API_VERSION.value]}/{domain}/messages",
                auth=(
                    "api",
                    messaging_config.secrets[
                        MessagingServiceSecrets.MAILGUN_API_KEY.value
                    ],
                ),
                data=data,
            )
            if not response.ok:
                logger.error(
                    "Email failed to send with status code: %s", response.status_code
                )
                raise MessageDispatchException(
                    f"Email failed to send with status code {response.status_code}"
                )
    except Exception as e:
        logger.error("Email failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Email failed to send due to: {Pii(e)}")


def _twilio_email_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: str,
) -> None:
    """Dispatches email using twilio sendgrid"""
    validate_config(messaging_config, "Twilio email")

    try:
        sg = sendgrid.SendGridAPIClient(
            api_key=messaging_config.secrets[
                MessagingServiceSecrets.TWILIO_API_KEY.value
            ]
        )

        # the pagination via the client actually doesn't work
        # in lieu of over-engineering this we can manually call
        # the next page if/when we hit the limit here
        response = sg.client.templates.get(
            query_params={"generations": "dynamic", "page_size": 200}
        )
        template_test = _get_template_id_if_exists(
            json.loads(response.body), EMAIL_TEMPLATE_NAME
        )

        from_email = Email(
            messaging_config.details[MessagingServiceDetails.TWILIO_EMAIL_FROM.value]
        )
        to_email = To(to.strip())
        subject = message.subject
        mail = _compose_twilio_mail(
            from_email, to_email, subject, message.body, template_test
        )

        response = sg.client.mail.send.post(request_body=mail.get())
        if response.status_code >= 400:
            logger.error(
                "Email failed to send: %s: %s",
                response.status_code,
                Pii(str(response.body)),
            )
            raise MessageDispatchException(
                f"Email failed to send: {response.status_code}, {Pii(str(response.body))}"
            )
    except Exception as e:
        logger.error("Email failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Email failed to send due to: {Pii(e)}")


def _twilio_sms_dispatcher(
    messaging_config: MessagingConfig,
    message: str,
    to: str,
) -> None:
    """Dispatches SMS using Twilio"""
    validate_config(messaging_config, "Twilio SMS", validate_details=False)

    account_sid = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_ACCOUNT_SID.value
    ]
    auth_token = messaging_config.secrets[
        MessagingServiceSecrets.TWILIO_AUTH_TOKEN.value
    ]
    messaging_service_id = messaging_config.secrets.get(
        MessagingServiceSecrets.TWILIO_MESSAGING_SERVICE_SID.value
    )
    sender_phone_number = messaging_config.secrets.get(
        MessagingServiceSecrets.TWILIO_SENDER_PHONE_NUMBER.value
    )

    client = Client(account_sid, auth_token)
    try:
        if messaging_service_id:
            client.messages.create(
                to=to, messaging_service_sid=messaging_service_id, body=message
            )
        elif sender_phone_number:
            client.messages.create(to=to, from_=sender_phone_number, body=message)
        else:
            logger.error(
                "Message failed to send. Either sender phone number or messaging service sid must be provided."
            )
            raise MessageDispatchException(
                "Message failed to send. Either sender phone number or messaging service sid must be provided."
            )
    except TwilioRestException as e:
        logger.error("Twilio SMS failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"Twilio SMS failed to send due to: {Pii(e)}")


def _aws_ses_dispatcher(
    messaging_config: MessagingConfig,
    message: EmailForActionType,
    to: str,
) -> None:
    validate_config(messaging_config, "AWS SES")

    aws_ses_serivce = AWS_SES_Service(messaging_config)

    try:
        aws_ses_serivce.send_email(to, message.subject, message.body)
    except Exception as e:
        logger.error("Email failed to send: {}", Pii(str(e)))
        raise MessageDispatchException(f"AWS SES email failed to send due to: {Pii(e)}")


def _get_template_id_if_exists(
    templates_response: Dict[str, List], template_name: str
) -> Optional[str]:
    """
    Checks to see if a SendGrid template exists for Fides, returning the id if so
    """

    for template in templates_response["result"]:
        if template["name"].lower() == template_name.lower():
            return template["id"]
    return None


def _compose_twilio_mail(
    from_email: Email,
    to_email: To,
    subject: str,
    message_body: str,
    template_test: Optional[str] = None,
) -> Mail:
    """
    Returns the Mail object to send, if a template is passed composes the Mail
    appropriately with the template ID and paramaterized message body.
    """
    if template_test:
        mail = Mail(from_email=from_email, subject=subject)
        mail.template_id = TemplateId(template_test)
        personalization = Personalization()
        personalization.dynamic_template_data = {"fides_email_body": message_body}
        personalization.add_email(to_email)
        mail.add_personalization(personalization)
    else:
        content = Content("text/html", message_body)
        mail = Mail(from_email, to_email, subject, content)
    return mail


def get_email_messaging_config_service_type(db: Session) -> Optional[str]:
    """
    Email connectors require that an email messaging service has been configured.
    Prefers Twilio if both Twilio email AND Mailgun has been configured.
    """

    # if there's a specified messaging service type, and it's an email service, we use that
    if (
        configured_service_type := ConfigProxy(
            db
        ).notifications.notification_service_type
    ) is not None and configured_service_type in EMAIL_MESSAGING_SERVICES:
        return configured_service_type

    # if no specified messaging service type, fall back to hardcoded preference hierarchy
    messaging_configs: Optional[List[MessagingConfig]] = MessagingConfig.query(
        db=db
    ).all()
    if not messaging_configs:
        # let messaging dispatch service handle non-existent service
        return None

    twilio_email_config = next(
        (
            config
            for config in messaging_configs
            if config.service_type == MessagingServiceType.twilio_email
        ),
        None,
    )
    if twilio_email_config:
        # First choice: use Twilio
        return MessagingServiceType.twilio_email.value

    mailgun_config = next(
        (
            config
            for config in messaging_configs
            if config.service_type == MessagingServiceType.mailgun
        ),
        None,
    )
    if mailgun_config:
        # Second choice: use Mailgun
        return MessagingServiceType.mailgun.value

    mailchimp_transactional_config = next(
        (
            config
            for config in messaging_configs
            if config.service_type == MessagingServiceType.mailchimp_transactional
        ),
        None,
    )
    if mailchimp_transactional_config:
        # Third choice: use Mailchimp Transactional
        return MessagingServiceType.mailchimp_transactional.value

    return None
